//
// cperiods.h -- header for periods and lattices of curves
//

#ifndef _CPERIODS_H_
#define _CPERIODS_H_

#include "curve.h"

#define TWOPI 2*Pi()
#define TWOPIEYE bigcomplex(0, TWOPI)


inline bigcomplex q(const bigcomplex& z) // q(z) = exp(2*pi*i * z)
{
  bigfloat twopix = TWOPI * real(z);
  return exp(-TWOPI * imag(z)) *
         bigcomplex( cos(twopix), sin(twopix) );
}



//
// the Cperiods class:
//

class Cperiods {
  bigcomplex w1, w2;
  bigcomplex tau;    // whenever meaningful, = w2/w1
  bigcomplex e1, e2, e3;  // 2-division values
  bigcomplex qtau, w1squared, w1cubed, sum3;
  int norm_code;  // 0 if not set, 
                  // 1 if normalized for lattice with Re(tau) = 0.5 "type1"
                  //   w1=2x real, w2=x+yi, lattice [2x, x+yi]
                  // 2 if "   "   "   "    "  "     "    "    = 0   "type2"
                  //   w1=x real, w2=yi pure imag, lattice [x,yi]
                  // 3 if normalized for tau in fundmental region
public:
  Cperiods() : w1(0), w2(0), tau(0), norm_code(0) {}
  Cperiods(bigcomplex ww1, bigcomplex ww2)  // nb must give two or no args
    : w1(ww1), w2(ww2), tau(w2/w1), norm_code(0)
    {}
  Cperiods(bigcomplex ww1, bigcomplex ww2, int normize) // normalize now per code
    : w1(ww1), w2(ww2), norm_code(0)
    { if (normize==1) norm_lattice();
      else norm_region(); }
  Cperiods(bigfloat x, bigfloat y, int type_set) // for known lattice types
    : norm_code(type_set)
    { if (type_set==1) {w1=2*x; w2=bigcomplex(x,y); }
      else {w1=x; w2=bigcomplex(0,y); }
      tau = w2/w1; }
  Cperiods(const Curve& E); // this will normalize for lattice, type=#comp's

  // copying:
  Cperiods(const Cperiods& cp)
    : w1(cp.w1), w2(cp.w2), tau(cp.tau), norm_code(cp.norm_code)
    {}
  void operator=(const Cperiods& cp)
    { w1=cp.w1; w2=cp.w2; tau=cp.tau; norm_code=cp.norm_code; }

  // normalizing:
  void norm_lattice(); // ie for w1 real and w2 pure Im or 2w2-w1 pure Im
                       // not yet implemented (and maybe never)
  void norm_region();  // ie for Im(t)>0, -.5<Re(t)<=.5, and |t| >= 1

  // member access functions
  friend inline bigcomplex gettau(const Cperiods& cp) {return cp.tau; }
     // up to caller to check norm_code
  friend inline int getnorm_code(const Cperiods& cp) {return cp.norm_code; }
     //0 unset, 1 lattice, 2 square lattice, 3 region
  int getwi(bigcomplex& ww1, bigcomplex& ww2) const
    { ww1=w1; ww2=w2; return norm_code; }

  // IO
  friend inline ostream& operator<<(ostream& os, const Cperiods& cp)
    { os<<"omega_1: "<<cp.w1<<"\tomega_2: "<<cp.w2
	<<"\ntau: "<<cp.tau<<"\tnorm: ";
      if (cp.norm_code==0) return os<<"none";
      if (cp.norm_code==1) return os<<"lattice type 1";
      if (cp.norm_code==2) return os<<"lattice type 2";
      return os<<"fundamental region";
    }

 // Weierstrass functions:
  bigcomplex X_coord(const bigcomplex& qz); // qz=q(z), z modulo lattice,
  bigcomplex Y_coord(const bigcomplex& qz); // gives coords in Y^2=4X^3+... model
  void XY_coords(bigcomplex& X, bigcomplex& Y, const bigcomplex& z);
};  // end of class Cperiods def'n


void makec4c6(Cperiods& cp, bigcomplex& cc4, bigcomplex& cc6);
   // NB this normalizes the arg cp; caller must save old copy if required

Curve trans_to_curve(Cperiods& cp);  // via makec4c6()


#endif

//ends file periods.h
