//
// curve.cc, adapted from elliptic.cc by OM by JEC with help from SCF
//

#include <stdio.h>     // for sprintf only, used in setting Kodaira symbols
#include <builtin.h>
#include "curve.h"

//
// IO operations
//

ostream& operator<<(ostream& os, const Curve& c)
{
  c.output(os);
  return os;
}

ostream& operator<<(ostream& os, const CurveRed& c)
{
  c.output(os);
  return os;
}

istream& operator>>(istream& is, Curve& c)
{
   c.input(is);
   return is ;
}


//Kraus' conditions:

int valid_invariants(const bigint& c4, const bigint& c6)  
{
  const bigint& disc = c4*c4*c4 - c6*c6;
  if (sign(disc)==0) return 0;    //
  if (ndiv(1728,disc)) return 0;  // need c4^3-c6^2=1728D, with D|=0
  long x6= mod(c6,27);
  if((x6==9)||(x6==-9)) return 0; // need c6 != +-9 (mod 27)
  x6 = mod(c6,4);
  if(x6==-1) return 1;            // OK if c6 = -1 (mod 4)
  if(ndiv(16,c4)) return 0;       // else need c4=0 (mod 16)
  x6=mod(c6,32);                  //      and
  return ((x6==0) || (x6==8));    //      c6 = 0,8 (mod 32).
}

void c4c6_to_ai(const bigint& c4, const bigint& c6, 
                bigint& a1, bigint& a2, bigint& a3, bigint& a4, 
                bigint& a6, 
                bigint& b2, bigint& b4, bigint& b6, bigint& b8)
{
  b2 = mod(-c6,12);
  const bigint& b22 = b2*b2;
  b4 = (b22-c4)/24;
  b6 = (-b2*b22+36*b2*b4-c6)/216;
  b8 = (b2*b6 - b4*b4) / 4;

  a1 = (odd(b2) ? 1 : 0);
  a3 = (odd(b6) ? 1 : 0);
  a2 = (b2-a1)/4;         // N.B. a1 == a1*a1 (= 0,1)
  a4 = (b4-a1*a3)/2;
  a6 = (b6-a3)/4;         // N.B. a3 == a3*a3 (= 0,1)
}

void c4c6_to_ai(const bigint& c4, const bigint& c6, 
                bigint& a1, bigint& a2, bigint& a3, bigint& a4, 
                bigint& a6)
{
  bigint b2, b4, b6, b8;
  c4c6_to_ai(c4,c6,a1,a2,a3,a4,a6,b2,b4,b6,b8);
}

void minimise_c4c6(const bigint& c4, const bigint& c6, const bigint& discr, 
                   bigint& newc4, bigint& newc6, bigint& newdiscr, bigint& u)
{
  u = 1; int u_is_1 = 1;
  newc4=c4; newc6=c6;
  const bigint& c62 = c6*c6;
  newdiscr = (c4*c4*c4-c62)/1728;
//  cout<<"At start of minimise_c4c6 with discr = " << discr << ", newdiscr = " << newdiscr << endl;
  const bigint& g = gcd( c62, newdiscr );
  const bigintArray& p_list = pdivs(g);
  bigintvar pr(p_list);
  bigint p;
  while ( pr.ok() )
  {
    p = pr.value(); pr++;
    long d = (long)floor(val(p,g)/12.0);
    if (p==2)
      {
	long a = I2long((c4 >> (4*d)) % 16);
	long b = I2long((c6 >> (6*d)) % 32); if(b<0) b+=32;
	if (( (b%4)!=3) && !( (a==0) && (( b==0) || (b==8) )))
	  {
	    d--;
	  }
      }
    else if (p==3) if (val(3,c6)==(6*d + 2)) d--;
    if(d>0) {u *= pow(p,d); u_is_1 = 0;}
  }
  if(!u_is_1)
    {
      bigint u2, u4, u6, u12;
      mul(u,u,u2); mul(u2,u2,u4); mul(u2,u4,u6); mul(u6,u6,u12);
      newc4 = c4 / u4;
      newc6 = c6 / u6;
      newdiscr /=  u12;
    }
//  cout<<"At end of minimise_c4c6 with discr = " << discr << ", newdiscr = " << newdiscr << endl;
}

//constructor for curve with invariants as argument
Curve::Curve(const bigint& c4, const bigint& c6)
{
  if (valid_invariants(c4, c6))
    {
      c4c6_to_ai(c4,c6,a1,a2,a3,a4,a6);
    }
  else 
    {
      cerr << " ## attempt to call Curve constructor\n"
           << "    with invalid invariants c4 = "<<c4<<", c6 = "<<c6
           << ": reading as null curve\n";
        a1=0; a2=0; a3=0; a4=0; a6=0;
    }
}


void Curve::output(ostream& os) const
    { os<<"["<<a1<<","<<a2<<","<<a3<<","<<a4<<","<<a6<<"]";}
//N.B. No spaces in output for ease of input to GP

void Curve::input(istream& is)
{
  is.flags( is.flags() | ios::dec );  //force decimal input (bug fix)
  char c;  // to eat commas and detect [ from {;
           // `{' flags curve input by invariants, eg {1,3}
           // `[' by coeffs a1--a6, eg [1,2,3,4,6]
           // seperators and terminators must then be present
           // (any nonnumeric will do after the first { or [ )
           // else assumes a1 a2 a3 a4 a6 separated by whitespace
  is>>c;   // This eats any whitespace first, and c is the next char input
//cout<<"First char read = "<<c<<"\n";
  switch (c) {
  case '{':
        {
//cout<<"Reading {c4,c6}...\n";
         bigint c4, c6;
         is >> c4 >> c >> c6 >> c;
         if (valid_invariants(c4, c6))
           {
             const bigint& b2 = mod(-c6,12);
             const bigint& b22 = b2*b2;
             const bigint& b4 = (b22-c4)/24;
             const bigint& b6 = (-b2*b22+36*b2*b4-c6)/216;
             a1 = (odd(b2) ? 1 : 0);
             a3 = (odd(b6) ? 1 : 0);
             a2 = (b2-a1*a1)/4;
             a4 = (b4-a1*a3)/2;
             a6 = (b6-a3*a3)/4;
           }
         else 
           {
//           cerr << " ## invalid invariants, reading as null curve\n";
             a1=0; a2=0; a3=0;a4=0; a6=0;
           }
          }
         break;
       case '[':
//cout<<"Reading [a1,a2,a3,a4,a6]...\n";
         is >> a1 >> c >> a2 >> c >> a3 >> c >> a4 >> c >> a6 >> c; 
         break;
       default:
         is.putback(c);
//cout<<"Put back that char and read as: a1 a2 a3 a4 a6 ...\n";
         is >> a1 >> a2 >> a3 >> a4 >> a6;
       }
}

// puts out TeX-ed equation of curve
void Curve::tex_print(ostream &os) const
{
        os << "$y^2" ;
        if(a1==0){
                ;
        } else {
                if(a1==1) os << " + xy" ;
                else if(a1==-1) os << " - xy" ;
                else if(a1 > 0) os << " +" << a1 << "xy" ;
                else os << " " << a1 << " xy" ;
        }
        if(a3==0){
                ;
        } else {
                if(a3==1) os << " + y" ;
                else if(a3==-1) os << " - y" ;
                else if(a3 > 0) os << " +" << a3 << "y" ;
                else os << " " << a3 << " y" ;
        }
        os << " = x^3" ;
        if(a2==0){
                ;
        } else {
                if(a2==1) os << " + x^2" ;
                else if(a2==-1) os << " - x^2" ;
                else if(a2 > 0) os << " +" << a2 << "x^2" ;
                else os << " " << a2 << " x^2" ;
        }
        if(a4==0){
                ;
        } else {
                if(a4==1) os << " + x" ;
                else if(a4==-1) os << " - x" ;
                else if(a4 > 0) os << " +" << a4 << "x" ;
                else os << " " << a4 << " x" ;
        }
        if(a6==0){
                ;
        } else {
                if(a6==1) os << " + 1" ;
                else if(a6==-1) os << " - 1" ;
                else if(a6 > 0) os << " +" << a6  ;
                else os << " " << a6 ;
        }
        os << "$" ;
        return ;
}


// end of file: curve.cc

//
// The following functions from OM's code are never used
//
// /* number of solutions to y^2+ay=b mod p */
// long quadroots(const bigint& a, const bigint& b, long p ) 
// {
//         if (p == 2) { 
//            if (!odd(a))  return(1) ; 
//            if (!odd(b))  return(2) ;
//            return(0) ;
//                 } 
//         else {
//          long d = I2long((a*a+4*b)%p);
//          if (d==0) return 1;
//          if (legendre(d,p)==1) return 2; 
//          return 0;
//               }
// }

// // for finding number of points mod 2 and 3
// long pointsmod(long p, const Curve& E)
// {
//   bigint a1,a2,a3,a4,a6;
//   E.getai(a1,a2,a3,a4,a6);
//   if (p == 2)
//     return( 1 + quadroots(a3, a6, 2) +
//            quadroots(a1 + a3, 1 + a2 + a4 + a6, 2)) ;
//   if (p ==3)
//     return( 1 + quadroots(a3, a6, 3) +
//            quadroots(a1 + a3, 1 + a2 + a4 + a6, 3)
//            + quadroots(-a1 + a3, -1 + a2 - a4 + a6, 3)) ;
//  // Now p>3
//   long count=0;
//   for (long x=0;x<p;x++) count+=quadroots(x*a1+a3,x*(x*(x+a2)+a4)+a6, p);
//   return  count;
// }



