//
// Curve.h, adapted from Elliptic.h by OM
//

// allow for multiple includes
#ifndef _ELLIPTIC_
#define _ELLIPTIC_

#include "marith.h"

class Curve; class Curvedata; class CurveRed;  class CurveRedArray;
class Point; class PointArray;
class IsogenyClass;

//general test:
int valid_invariants(const bigint& c4, const bigint& c6); //true if valid
void c4c6_to_ai(const bigint& c4, const bigint& c6, 
		bigint& a1, bigint& a2, bigint& a3, bigint& a4, 
		bigint& a6, 
		bigint& b2, bigint& b4, bigint& b6, bigint& b8);
void c4c6_to_ai(const bigint& c4, const bigint& c6, 
		bigint& a1, bigint& a2, bigint& a3, bigint& a4, 
		bigint& a6);
void minimise_c4c6(const bigint& c4, const bigint& c6, const bigint& discr, 
                   bigint& newc4, bigint& newc6, bigint& newdiscr, bigint& u);

//
// base class for bare elliptic curve
//

class Curve{ 
friend class Point;
friend class IsogenyClass;
public:
  int isnull() const
    {return ((a1==0)&&(a2==0)&&(a3==0)&&(a4==0)&&(a6==0));}
  void getai(bigint& aa1, bigint& aa2, bigint& aa3,
             bigint& aa4, bigint& aa6) const
    {aa1=a1; aa2=a2; aa3=a3; aa4=a4; aa6=a6; }
    
// input and output
  void output(ostream& os) const;
  void input(istream& is);
  void tex_print(ostream &) const ;
  // puts out TeX-ed equation of curve; never been used

// constructors 
  Curve(void) //  :a1(0),a2(0),a3(0),a4(0),a6(0) 
    {a1=0;a2=0;a3=0;a4=0;a6=0;}
  Curve(const bigint& c4, const bigint& c6); //init by invariants
                                 //check valid for elliptic curve
                                 //if not create null curve and cerr
  Curve(const bigint& aa1, const bigint& aa2, const bigint& aa3,
        const bigint& aa4, const bigint& aa6) 
    :a1(aa1),a2(aa2),a3(aa3),a4(aa4),a6(aa6) {}
  Curve(const Curve& c)
    : a1(c.a1), a2(c.a2), a3(c.a3), a4(c.a4), a6(c.a6)
      {}
  void operator=(const Curve& c)
    { a1=c.a1; a2=c.a2; a3=c.a3; a4=c.a4; a6=c.a6; }
// no destructor is needed

// equality tests -- inherited, so you can compare Curvedata with Curve
  int operator==(const Curve& f) const
    {return ((a1==f.a1)&&(a2==f.a2)&&
             (a3==f.a3)&&(a4==f.a4)&&(a6==f.a6));}
  int operator!=(const Curve& f) const
    {return ((a1!=f.a1)||(a2!=f.a2)||
             (a3!=f.a3)||(a4!=f.a4)||(a6!=f.a6));}
protected:
  bigint a1 ;
  bigint a2 ;
  bigint a3 ;
  bigint a4 ;
  bigint a6 ;
} ;


//
// derived class for curve with computed invariants
//  NB minimalize() is a member function, flagged if ever called; by
//  default not called, but see constructors for how to force.
//

class Curvedata : public Curve{
friend class CurveRed;  // bug fix; ridiculous, is a derived class
friend class IsogenyClass;
public:
  Curvedata();
  Curvedata(const bigint& aa1, const bigint& aa2, const bigint& aa3,
        const bigint& aa4, const bigint& aa6, int min_on_init = 0);
  Curvedata(const Curve& c, int min_on_init = 0);
  Curvedata(const bigint& cc4, const bigint& cc6, int min_on_init = 0);
  Curvedata(const Curvedata& c, int min_on_init = 0);
       // nb compiler cannot generate copy because constructor from
       // curve overrides.
       // But default assign, destruct suffice
  void operator=(const Curvedata& c);
  void minimalize();  // Changes self in situ
  Curvedata minimalize(bigint& u, bigint& r, bigint& s, bigint& t) const;
     // Self unchanged and returns transformation
  void transform(const bigint& r, const bigint& s, const bigint& t);   // NB  u = 1;
                            // the more general case is not implemented here
  void output(ostream& os) const;
  void input(istream& is);
  PointArray* make_torsion();  // implemented in points.cc
  inline long get_ntorsion() 
    {
      if(ntorsion==0) make_torsion();
      return ntorsion; 
    }
  void getbi(bigint& bb2, bigint& bb4, bigint& bb6, bigint& bb8) const
    {bb2=b2; bb4=b4; bb6=b6; bb8=b8; }
  void getci(bigint& cc4, bigint& cc6) const
    {cc4=c4; cc6=c6; }
  friend inline bigint getb2(const Curvedata& c) {return c.b2; }
  friend inline bigint getb4(const Curvedata& c) {return c.b4; }
  friend inline bigint getb6(const Curvedata& c) {return c.b6; }
  friend inline bigint getb8(const Curvedata& c) {return c.b8; }
  friend inline bigint getc4(const Curvedata& c) {return c.c4; }
  friend inline bigint getc6(const Curvedata& c) {return c.c6; }
  friend inline bigint getdiscr(const Curvedata& c) {return c.discr; }
  friend inline int getconncomp(const Curvedata& c) {return c.conncomp; }
  friend inline bigintArray getbad_primes(const Curvedata& c)
    {return c.bad_primes; }
protected:
  bigint b2 ;
  bigint b4 ;
  bigint b6 ;
  bigint b8 ;
  bigint c4 ;
  bigint c6 ;
  bigint discr ;
  int minimal_flag;  // 0 if .minimalize() has not been called
  bigintArray bad_primes; //prime divisors of discriminant
  int conncomp ;    // number of components (1 or 2)
  long ntorsion; // 0 if .gettorsion() not called
} ;

//
// further derived class for curve, invariants and periods:
//

class CurvedataExtra : public Curvedata{ 
public:
  CurvedataExtra(const Curvedata&) ; 
  virtual void output(ostream& os) const;
  void input(istream& is)
    {cerr<<"*** You cannot input a CurvedataExtra -- must be just Curve\n";
     exit(1); }
  void getroots(bigfloat& r1, bigfloat& r2, bigfloat& r3) const
    {r1=roots[0]; r2=roots[1]; r3=roots[2]; }
      // NB caller should then look at conncomp to see how many are set
  friend inline bigfloat getperiod(const CurvedataExtra c) {return c.period; }
protected:
  bigfloat roots[3] ;     // real two-division points; NB if there's only one,
                        // it is stored in roots[2]
  bigfloat period ;       // smallest real period * conncomp
                
} ;

//
// CurveRed class call Tates algorithm as constructor,
// stores the info as member variables, using sub-structures as follows
// (client need know nothing about these; could hide)

// class Kodaira_code just holds an int which "codes" the type as follows:
// (this coding originally from R.G.E.Pinch)
//
// Im                 -> 10*m
// I*m                -> 10*m+1
// I, II, III, IV     -> 1, 2, 3, 4
// I*, II*. III*, IV* -> 5, 6, 7, 8
//
class  Kodaira_code {
public:
  int code;
//
  Kodaira_code(int k = -1) : code(k) {;}
  Kodaira_code(const Kodaira_code& c) : code(c.code) {;}
  void operator=(int k) {code=k;}
  void operator=(const Kodaira_code& c) {code=c.code;}
};

ostream& operator<<(ostream& os, const Kodaira_code& c);


class Reduction_type {
public:
  int ord_p_discr;
  int ord_p_N;
  int ord_p_j_denom;
  Kodaira_code Kcode;  // NB the constructor makes this from an int
  int c_p;
//
  Reduction_type(int opd=0, int opN=0, int opj=0, int kc=-1, int cp=0)
    : ord_p_discr(opd), ord_p_N(opN), ord_p_j_denom(opj), Kcode(kc), c_p(cp)
    {}
  void operator=(const Reduction_type& r)
    {ord_p_discr=r.ord_p_discr;ord_p_N=r.ord_p_N;ord_p_j_denom=r.ord_p_j_denom;
     Kcode=r.Kcode;c_p=r.c_p; }
  friend inline ostream& operator<<(ostream& os, const Reduction_type& R);
};


class CurveRed : public Curvedata {
friend class IsogenyClass;
protected:
  Reduction_type* reduct_array;   //just a plain c++ array
  bigint N;                      //the conductor
public:
  CurveRed() : Curvedata(), reduct_array(0) {N=0;}
  CurveRed(const Curvedata& E);  // construct by Tate's algorithm
             // arg E need not be minimal, but the reduced form will be
  ~CurveRed();
  CurveRed(const CurveRed& E);
  void operator=(const CurveRed& E);
  void display(ostream& os) const; // full output
  void output(ostream& os) const;  // just the curvedata and conductor

  // member access functions:
  friend inline bigint getconductor(const CurveRed& c) {return c.N; }
  friend int getord_p_discr(const CurveRed& c, const bigint& p);
  friend int getord_p_N(const CurveRed& c, const bigint& p);
  friend int getord_p_j_denom(const CurveRed& c, const bigint& p);
  friend int getc_p(const CurveRed& c, const bigint& p);
  friend int prodcp(const CurveRed& c);
  friend Kodaira_code getKodaira_code(const CurveRed& c, const bigint& p);
    // the returned value casts as a character array; to use coded as int,
    // say declared Kodaira_code Kc, just use public member Kc.code
};



ostream& operator<<(ostream& os, const Curve& c);
ostream& operator<<(ostream& os, const Curvedata& c);
ostream& operator<<(ostream& os, const CurvedataExtra& c);
ostream& operator<<(ostream& os, const CurveRed& c);
istream& operator>>(istream& is, Curve& c);
istream& operator>>(istream& is, Curvedata& c);  
   //Reads the curve (ai) and computes the rest.


//
// CurveRedArrays and CurveRedvars : copy of longlists, longvars
//

class CurveRedArray {
friend class CurveRedvar;
  long length;        
  CurveRed *items;  
public:
  CurveRedArray(long n=0);
  ~CurveRedArray();
  CurveRedArray(const CurveRedArray& l); 
  CurveRedArray(const CurveRedArray& l, long n);
  CurveRedArray& operator=(const CurveRedArray&l);
  CurveRed& get(long i);
  long getlength() {return length;}
  long locate(const CurveRed& num) const;
                              //return index if in list, else -1
  CurveRed& operator[](long i) {return get(i);}
  CurveRed operator()(long i) const;
  void set(long i, CurveRed val);
  void truncate(long newlength) 
    { if (newlength<length) length = newlength;}
      // no need to deallocate memory here because copy constructors do so
      // almost always it is the temporary of a function that calls truncate
  void output(ostream& os) const;
  friend inline ostream& operator<<(ostream& os, const CurveRedArray& l)
    {l.output(os);return os;}
};

class CurveRedvar {
public:
  long index;        /* current index */
private:
  CurveRed* values;
  long maxindex;     /* max index */
public:
  CurveRedvar(const CurveRedArray& l) 
    {maxindex=l.length-1; index=0; values=l.items;}
  void init(const CurveRedArray& l) 
    {maxindex=l.length-1; index=0; values=l.items;}
  int next() {if ((index++)<maxindex) {values++; return 1;}
              else return 0;}
  CurveRed operator++()
    {if ((index++)<=maxindex) {CurveRed ans=*values;values++; return ans;}
     else return CurveRed();}
  int ok() const {return index<=maxindex;}
  int more() const {return index<maxindex;}
  CurveRed value() const {return *values;}
//  operator CurveRed() const {return *values;}
};



// end of file: curve.h

#endif
