//
// curvered.cc -- implementation of CurveRed class etc.
//

#include "curve.h"

const bigint m1(1);
const bigint m2(2);
const bigint m3(3);

ostream& operator<<(ostream& os, const Kodaira_code& c)
{
  int code=c.code;
  switch (code%10){
  case 0:
    os<<"I"<<((code)/10); break;
  case 1:
    os<<"I*"<<((code - 1)/10);  break;
  case 2:
    os<<"II   "; break;
  case 3:
    os<<"III  "; break;
  case 4:
    os<<"IV   "; break;
  case 5:
    os<<"IV*  "; break;
  case 6:
    os<<"III* "; break;
  case 7:
    os<<"II*  "; break;
  default:
    os<<"???? "; break;
  };
}

//
// Tate's algorithm -- constructor function for a CurveRed
//
// subroutines -- not general purpose (.cc only)

bigint root(const bigint& aa, int e, const bigint & p)  
// the e'th root of aa, mod p
{  
  bigint ans, temp; int found=0;
  const bigint& a = mod(aa, p);
  for (bigint i = 1; (! found); ++i)
    {ans = i;
     if (e==2) temp = ans*ans - a;
     else temp = ans*ans*ans - a;
     found = div(p, temp); }
//   bigint ans;
//   int found = 0;
//   bigint a = aa % p;
//   for ( ans = 1; (ans<=p) && (! found); ++ans)
//     found = div(p, pow(ans,e) - a);     //N.B. pow(0,p) causes problems.
  return ans;
}

// test if quadratic aX^2 + bX + c = 0 (mod p) has roots
int rootsexist(const bigint& aa, const bigint& bb, const bigint& cc, const bigint& p)
{
  const bigint& a = aa % p;
  const bigint& b = bb % p;
  const bigint& c = cc % p;
  const bigint& temp = (a*b*c)%p;
  if (p==m2) return sign(temp)==0;
  if (sign(a)==0) return 1;
  const bigint& d = b*b - 4*a*c;
  return legendre(d, p) >= 0;      //ie true if legendre(d,p)=0,1
}

// find the number of roots of X^3 + bX^2 + cX + d = 0 (mod p)
int nrootscubic(const bigint& bb, const bigint& cc, const bigint& dd, const bigint& p)
{
//cout<<"nrootscubic called with p="<<p<<" and b, c, d = "<<bb<<", "<<cc<<", "<<dd<<endl;
  const bigint& b = mod(bb, p);
  const bigint& c = mod(cc, p);
  const bigint& d = mod(dd, p);
//cout<<"which reduce to b, c, d = "<<b<<", "<<c<<", "<<d<<endl;
  bigint root1;
  int found = 0;
  for (bigint i = 0; i < p ; ++i)
    {
      bigint temp = (((i+b)*i + c)*i + d)%p;
      if (sign(temp)==0 ) {root1 = i; found = 1; break;} 
    }
//cout<<"Root search finds returns " << found;
//if(found)cout<<" (root = "<<i<<") ";
//cout<<endl;
  if (found)
    {const bigint& e = bb + root1;
     const bigint& f = cc + e*root1;
     if ( rootsexist(m1,e,f,p) ) return 3;
     else return 1; }
  else return 0;
}

CurveRed::~CurveRed() 
{
  delete[] reduct_array; 
}

CurveRed::CurveRed(const CurveRed& E)    : Curvedata(E), N(E.N)
{ 
  reduct_array = new Reduction_type[bad_primes.length];
  for (int index=0; index<bad_primes.length; index++)
    reduct_array[index]=E.reduct_array[index]; 
}

void CurveRed::operator=(const CurveRed& E)
{
  if(this==&E) return;
  a1=E.a1;a2=E.a2;a3=E.a3;a4=E.a4;a6=E.a6;b2=E.b2;b4=E.b4;b6=E.b6;b8=E.b8;
  c4=E.c4;c6=E.c6;discr=E.discr;minimal_flag=1;bad_primes=E.bad_primes;
  conncomp=E.conncomp;
  ntorsion=E.ntorsion;
  delete[] reduct_array;
  reduct_array = new Reduction_type[bad_primes.length];
  for (int index=0; index<bad_primes.length; index++)
    {reduct_array[index]=E.reduct_array[index]; }
  N=E.N; 
}

CurveRed::CurveRed(const Curvedata& E)
: Curvedata(E, 1), N(1) //minimalize in constructor
{
  // constructor stuff
  if (discr==0) {N = 0; reduct_array = 0; return; }
  reduct_array = new Reduction_type[bad_primes.length];

  // local variables
  Curvedata C(*this);
//  C.a1=a1;C.a2=a2;C.a3=a3;C.a4=a4;C.a6=a6;C.b2=b2;C.b4=b4;C.b6=b6;C,b8=b8;
//  Curvedata C(E); // local copy subject to transformations
//  C.minimalize(); // nb since we always have u=1, c4,c6 and discr never change
  bigint p, halfmodp, temp, r, s, t, b, c, bb, cc, bc, d, w, x, mx, my,
          a2t, a3t, a4t, a6t;
  int ord_p_discr, ord_p_N, ord_p_j, c_p, pdiv2, pdiv3, sw, loop, ix, iy;

  // main loop - for each of the prime divisors of the discriminant.
  // Because the curve is minimal, Tate's algorithm reduce-loop is not needed

//cout<<"Running Tate's algorithm"<<endl;
  int index = 0;
  for (bigintvar pvar(bad_primes);
       pvar.ok();
       N *= pow(p, reduct_array[index].ord_p_N), index++)
  {
    p = pvar.value(); pvar++;
//cout<<"p = " << p << endl;
    ord_p_discr = val(p,discr);
    ord_p_j = ord_p_discr - 3*val(p,c4);
    if (ord_p_j < 0) ord_p_j = 0;
    halfmodp = (m1 + p) / m2;
    pdiv2 = (p==m2);
    pdiv3 = (p==m3);

    //change coords so that p|C.a3,C.a4,C.a6
    if ( pdiv2 )
      { if ( div(p,C.b2) )
          { r = root(C.a4,2,p);
            t = root(((r+C.a2)*r+C.a4)*r+C.a6,2,p); }
        else { temp=invmod(C.a1,p);
               r = temp*C.a3;
               t = temp*(C.a4 + r*r); }
      }
    else if ( pdiv3 )
      { if ( div(p,C.b2) ) r = root(-C.b6,3,p); else r = -invmod(C.b2,p)*C.b4;
        t  =  C.a1*r + C.a3; }
    else
      { if ( div(p,c4) ) r = -invmod(12,p)*C.b2;
        else r = -invmod(12*c4,p)*(c6+C.b2*c4);
        t =  -halfmodp*(C.a1*r+C.a3); }
    r = mod(r,p);
    t = mod(t,p);
    C.transform(r,0,t);
    
    // test for Types In, II, III, IV
    if ( ndiv(p,c4) )
      {temp=-C.a2;
       if (rootsexist(m1,C.a1,temp,p) ) c_p = ord_p_discr;
       else { if ( odd(ord_p_discr) ) c_p = 1; else c_p = 2; }
       reduct_array[index] = Reduction_type
         (ord_p_discr, 1, ord_p_j, 10*ord_p_discr, c_p);
       continue; }  // Type In (n=ord_p_discr)
    else if ( val(p,C.a6) < 2 )
      {reduct_array[index] = Reduction_type
         (ord_p_discr, ord_p_discr, ord_p_j, 2, 1);
       continue; } // Type II
    else if ( val(p,C.b8) < 3 )
      {reduct_array[index] = Reduction_type
         (ord_p_discr, ord_p_discr - 1, ord_p_j, 3, 2);
       continue; } // Type III
    else if ( val(p,C.b6) < 3 )
      {temp = -(C.a6/p)/p;
       bigint temp2 = C.a3/p;
       if ( rootsexist(m1,temp2,temp,p) ) c_p = 3;
      else c_p = 1;
       reduct_array[index] = Reduction_type
         (ord_p_discr, ord_p_discr - 2, ord_p_j, 4, c_p);
       continue; } // Type IV
     
    // else change coords so that p|C.a1,C.a2, p^2|C.a3,C.a4, p^3|C.a6
    if ( pdiv2 )
      { s = root(C.a2,2,p);
        t = p*root((C.a6/p)/p,2,p); }
    else if ( pdiv3 )
      { s = C.a1;
        t = C.a3; }
    else
      { s = -C.a1*halfmodp;
        t = -C.a3*halfmodp; }
    C.transform(0,s,t);

    //                             3     2
    // Analyse roots of the cubic T  + bT  + cT + d = 0, where
    // b=C.a2/p, c=(C.a4/p)/p, d=((C.a6/p)/p)/p
    b=C.a2/p;
    c=(C.a4/p)/p;
    d=((C.a6/p)/p)/p;
    bb=b*b; cc=c*c; bc=b*c;
    w = 27*d*d - bb*cc + 4*b*bb*d - 18*bc*d + 4*c*cc;
    x = 3*c - bb;
       
    if ( div(p,w) )
      {if ( div(p,x) ) sw = 3; else sw = 2; }
    else sw = 1;
  
//cout << "Analusing roots of cubic; case " << sw << endl;

    switch ( sw ) {
    case 1:
      //Three distinct roots - Type I*0
      reduct_array[index] = Reduction_type
        (ord_p_discr, ord_p_discr - 4, ord_p_j, 1, 1+nrootscubic(b,c,d,p) );
      break;

    case 2:
      // One double root - Type I*m for some m
      // Change coords so that the double root is T=0 mod p
      if ( pdiv2 ) r = root(c,2,p);
      else if ( pdiv3 ) r = c*invmod(b,p);
      else r = (bc - 9*d)*invmod(2*x,p);
      r = p * mod(r,p);
      C.transform(r,0,0);

      ix = 3; iy = 3; mx = p*p; my = p*p;
      loop = 1;
      while (loop)
      {
        a2t = C.a2/p;
        a3t = C.a3/my;
        a4t = (C.a4/p)/mx;
        a6t = (C.a6/mx)/my;
	temp = a3t*a3t + 4*a6t; 
        if ( div(p,temp ) )
          {if ( pdiv2 ) t = my*root(a6t,2,p);
           else t = my*mod(-a3t*halfmodp, p);
           C.transform(0,0,t);
           my = my*p;
           iy++;
           a2t = C.a2/p;
           a3t = C.a3/my;
           a4t = (C.a4/p)/mx;
           a6t = (C.a6/mx)/my;
	   temp = a4t*a4t - 4*a6t*a2t;
           if ( div(p,temp) )
             {if ( pdiv2 ) r = mx*root( a6t*invmod(a2t,p), 2, p);
              else r = mx*mod( -a4t*invmod(2*a2t,p), p);
              C.transform(r,0,0);
              mx = mx*p;
              ix++;       // and stay in loop
             }
           else
             {if ( rootsexist(a2t,a4t,a6t,p) ) c_p = 4;
              else c_p = 2;
              loop = 0; }  // and exit loop
          }
        else
          { temp = -a6t;
	    if ( rootsexist(m1,a3t,temp,p) ) c_p = 4;
           else c_p = 2;
           loop = 0; }
      }
      reduct_array[index] = Reduction_type
        (ord_p_discr, ord_p_discr - ix - iy + 1, ord_p_j,
         10 * (ix + iy) - 49, c_p );
      break;  // Type I*m
    
    case 3:
      // Triple root
      // change coords so that T=0 mod p
      if ( pdiv2 ) r = b;
      else if ( pdiv3 ) r = root(-d,3,p);
      else r = -b*invmod(3,p);
      r = p*mod(r,p);
      C.transform(r,0,0);

      a3t = (C.a3/p)/p;
      a6t = (((C.a6/p)/p)/p)/p;

      // test for Type IV*
      temp = a3t*a3t + 4*a6t;
      if ( ndiv(p,temp ) )
        {
	  temp = -a6t;
	  if ( rootsexist(m1,a3t,temp,p) ) c_p = 3;
	  else c_p = 1;
	  reduct_array[index] = Reduction_type
	    (ord_p_discr, ord_p_discr - 6, ord_p_j, 5, c_p);
	  break; }
      
      // change coords so that p^3|C.a3, p^5|C.a6
      if ( pdiv2 ) t = -p*p*root(a6t,2,p);
      else t = p*p*mod(-a3t*halfmodp, p);
      C.transform(0,0,t);

      // test for types III*, II*
      if ( val(p,C.a4) < 4 )
        {reduct_array[index] = Reduction_type
           (ord_p_discr, ord_p_discr - 7, ord_p_j, 6, 2);
         break; }  // Type III*
      else if ( val(p,C.a6) < 6 )
        {reduct_array[index] = Reduction_type
           (ord_p_discr, ord_p_discr - 8, ord_p_j, 7, 1);
         break; } // Type II*

      else
        cerr<<" ## Tate's algorithm reached end of loop !!!"<<endl;

      // at this point (only if the input curve were not minimal)
      // one would divide each ai by p^i and start again
    };  // end switch
  }     // end primes for-loop
  return;
}     // end of Tate's algorithm


// CurveRed member access friend functions:

int getord_p_discr(const CurveRed& c, const bigint& p)
{
  int index = c.bad_primes.locate(p);
  if (index < 0) return 0;
  return c.reduct_array[index].ord_p_discr;
}

int getord_p_N(const CurveRed& c, const bigint& p)
{
  int index = c.bad_primes.locate(p);
  if (index < 0) return 0;
  return c.reduct_array[index].ord_p_N;
}

int getord_p_j_denom(const CurveRed& c, const bigint& p)
{
  int index = c.bad_primes.locate(p);
  if (index < 0) return 0;
  return c.reduct_array[index].ord_p_j_denom;
}

int getc_p(const CurveRed& c, const bigint& p)
{
  int index = c.bad_primes.locate(p);
  if (index < 0) return 1;
  return c.reduct_array[index].c_p;
}

int prodcp(const CurveRed& c)
{
  int ans=1;
  for(int i=0; i<c.bad_primes.length; i++) ans*=c.reduct_array[i].c_p;
  return ans;
}

Kodaira_code getKodaira_code(const CurveRed& c, const bigint& p)
{
  int index = c.bad_primes.locate(p);
  if (index < 0) return Kodaira_code(0);  // ie good reduction
  return c.reduct_array[index].Kcode;
}

void CurveRed::output(ostream& os) const
{
  Curvedata::output(os);
  if(isnull()) return;
  os << "Conductor = " << N << endl;
}

ostream& operator<<(ostream& os, const Reduction_type& R)
{
  os << R.ord_p_discr << "\t";
  os << R.ord_p_N << "\t";
  os << R.ord_p_j_denom <<"\t";
  os << R.Kcode << "\t";
  os << R.c_p;
  return os;
}

void CurveRed::display(ostream& os) const
{
  CurveRed::output(os);
  if(isnull()) return;
  os << "Reduction type at bad primes:\n";
  os <<"p\tord(d)\tord(N)\tord(j)\tKodaira\tc_p\n";
  for (int i = 0; i<bad_primes.length; i++)
      os << bad_primes(i) << "\t" << reduct_array[i] << endl;
}

// CurveRedArray and CurveRedvar functions

CurveRedArray::CurveRedArray(long n) 
{
  length=n; items=new CurveRed[n];
}

CurveRedArray::~CurveRedArray() 
{
  delete[] items; length = 0; 
}

CurveRedArray::CurveRedArray(const CurveRedArray& l) 
{
  length=l.length;
  items=new CurveRed[length]; 
  long n=length; CurveRed* x=items, *y=l.items;
  while(n--) {*x++=*y++;} 
}

CurveRedArray::CurveRedArray(const CurveRedArray& l, long n) 
{
  length= n<=l.length ? n : l.length;
  items=new CurveRed[length]; 
  n=length; CurveRed* x=items, *y=l.items;
  while(n--) {*x++=*y++;} 
}

CurveRedArray& CurveRedArray::operator=(const CurveRedArray&l)
{
  if (this==&l) return *this;
  delete[] items;
  length=l.length;
  items=new CurveRed[length]; 
  long n=length; CurveRed* x=items, *y=l.items;
  while(n--) {*x++=*y++;}
  return *this; 
}

CurveRed& CurveRedArray::get(long i)
{
  if ((i>=0)&&(i<length)) return items[i];
  else 
    {
      cerr<<"bad index "<<i<<" in CurveRedArray!\n"; return items[0];
    } 
}

CurveRed CurveRedArray::operator()(long i) const
{
  if ((i>=0)&&(i<length)) return items[i]; 
  else{cerr<<"bad index "<<i<<" in CurveRedArray!\n";
       return CurveRed();
     }
}

void CurveRedArray::set(long i, CurveRed val)
{
  if ((i>=0)&&(i<length)) items[i]=val; 
  else {cerr<<"bad index "<<i<<" in CurveRedArray!\n";}
}

void CurveRedArray::output(ostream& os) const
{
  long i;CurveRed *it;
  for(i=0,it=items;i<length;i++,it++)
    {if(i)os<<", ";os<<(Curve)(*it);} 
}
  
long CurveRedArray::locate(const CurveRed& num) const
{
  for ( int i=0; i<length; i++) if (num == items[i]) return i;
  return -1;
}
