// interface.h: provides common interface for LiDIA or G++

//  The following macro switches can be used: it is best to define them
//  in Makefiles rather than changing this file.
//  1. LiDIA_INTS   Use LiDIA for bigints and rationals only
//  2. LiDIA        Use LiDIA also for bifloats and bigcomplexes
//  3. (neither)    Don't use LiDIA at all: use libg++ for bigints (Integers)

#ifndef _INTERFACE_H_
#define _INTERFACE_H_

#ifdef LiDIA
#define LiDIA_INTS
#endif

#include <builtin.h>

// integers and rationals

//////////////////////////
#if defined(LiDIA_INTS) 
//////////////////////////

#include "LiDIA/bigint.h"

// Thomas Papanikolaou: LiDIA has these functions declared after version 1.3.0
#if (LIDIA_MAJOR_VERSION <= 1 && LIDIA_MINOR_VERSION <= 3 && LIDIA_PATCH_VERSION < 1) 
inline int is_zero(const bigint& x) {return x.is_zero();}
inline int is_positive(const bigint& x) {return x.is_positive();}
inline int is_negative(const bigint& x) {return x.is_negative();}
inline int is_one(const bigint& x) {return x.is_one();}
#endif

inline int odd(const bigint& a) {return a.is_odd();}
inline int even(const bigint& a) {return a.is_even();}
inline long lg(const bigint& x) {return x.bit_length()-1;}
inline void longasI(long& a, bigint x) {x.longify(a);}
inline int sign(const bigint& a) {return a.sign();}
inline long I2long(const bigint& x) {long a; x.longify(a); return a;}
inline double I2double(const bigint& x) { return dbl(x);}
inline void negate(bigint& a) {a.negate();}
inline bigint sqrt(const bigint& a) 
{bigint b(0); sqrt(b,a); return b;}
inline bigint sqr(const bigint& a) {bigint b; square(b,a); return b;}
//inline bigint lcm(const bigint& a, const bigint& b)
// {if (a.is_zero() && b.is_zero()) return 0; return a*(b/gcd(a,b));}
inline bigint atoI(char* s) {bigint a; string_to_bigint(s,a); return a;}
inline void addx(const bigint& a, const bigint& b, bigint& c) {add(c,a,b);}
inline void sub(const bigint& a, const bigint& b, bigint& c) {subtract(c,a,b);}
inline void mul(const bigint& a, const bigint& b, bigint& c) {multiply(c,a,b);}
inline void div(const bigint& a, const bigint& b, bigint& c) {divide(c,a,b);}
inline void rshift(bigint& a, long i, bigint& c) {shift_right(c,a,i);}
inline void lshift(bigint& a, long i, bigint& c) {shift_left(c,a,i);}
inline bigint pow(const bigint& a, const bigint& e) 
 {bigint b; power(b,a,e); return b;}
inline bigint pow(const bigint& a, long e) 
 {bigint b; power(b,a,e); return b;}

#include "LiDIA/bigrational.h"

//////////////////////////////////////////////////////////////////
#else  // non-LiDIA (libg++) Integers and Rationals
//////////////////////////////////////////////////////////////////

#include <Integer.h>
#define bigint Integer
#define bigrational Rational
#define power pow

// integers

inline int is_zero(const bigint& x) {return sign(x)==0;}
inline int is_positive(const bigint& x) {return sign(x)>0;}
inline int is_negative(const bigint& x) {return sign(x)<0;}
inline int is_one(const bigint& x) {return sign(x-1)==0;}
inline void longasI(long& a, const bigint& x) {a = x.as_long();}
inline int is_long(const bigint& a) {return a.fits_in_long();}
inline void negate(bigint& a) {a.negate();}
inline void sqrt(bigint& a, const bigint& b) {a=sqrt(b);}
inline void square(bigint& a, const bigint& b) {a=sqr(b);}
// for LiDIA-compatibility: 
inline void addx(const bigint& a, const bigint& b, bigint& c) {add(a,b,c);}
inline void swap(bigint& a, bigint& b) {bigint c(a); a=b; b=c;}

// rationals

#include <Rational.h>

#endif // Integers and rationals

// Reals and Complexes

//////////////////////////////////////////////////////////////////
#ifdef LiDIA
//////////////////////////////////////////////////////////////////

// reals

#include "LiDIA/bigfloat.h"

inline long decimal_precision() {return bigfloat::get_precision();}
inline int is_zero(bigfloat x) {return x.is_zero();}
inline int is_approx_zero(bigfloat x) {return x.is_approx_zero();}
inline void Iasb(long& a, bigfloat x) {x.longify(a);}
inline long longify(bigfloat x) {long a; x.longify(a); return a;}
inline bigfloat I2bigfloat(const bigint& x) { bigfloat y(x); return y;}

// complexes

#include "LiDIA/bigcomplex.h"
#define Complex bigcomplex

inline int is_zero(const Complex& z) {return z.is_zero();}
inline int is_approx_zero(const Complex& z) 
  {return z.real().is_approx_zero()&&z.imag().is_approx_zero();}
//  {return abs(z).is_approx_zero();}
inline bigcomplex pow(const bigcomplex& a, const bigcomplex& e) 
 {return power(a,e);}
inline bigcomplex pow(const bigcomplex& a, const bigfloat& e) 
 {return power(a,e);}
inline bigcomplex pow(const bigcomplex& a, long e) 
 {return power(a,e);}
inline bigfloat pow(const bigfloat& a, const bigfloat& e) 
 {return power(a,e);}
inline bigfloat pow(const bigfloat& a, long e) 
 {return power(a,e);}

//////////////////////////////////////////////////////////////////
#else  // non-LiDIA doubles and libg++ Complexes
//////////////////////////////////////////////////////////////////

// reals

#define bigfloat double
#define bigcomplex Complex

inline long decimal_precision() {return 15;}
inline int is_zero(double x) {return abs(x)<1e-15;}
inline int is_approx_zero(double x) {return abs(x)<1e-10;}
#ifndef LIDIA_BIGFLOAT_H
inline double Pi(void) {return M_PI;}
inline double Euler(void) {return 0.577215664901532860606512;}
#endif
inline double round(double x) {return floor(x+0.5);}
inline void Iasb(long& a, double x) {a = (long)x;}
inline long longify(double x) {return (long)x;}

#ifndef LiDIA_INTS   // Then we are using libg++ Integers
#ifdef OLD_CONVERSIONS
#define I2double(x) (double)((x))      //Old
#define I2long(x) (long)((x))          //Old
#else
#define I2double(x) (x).as_double()  //New
#define I2long(x) (x).as_long()      //New
#endif

#endif  // of #ifndef LiDIA_INTS

inline bigfloat I2bigfloat(const bigint& x) {return I2double(x);}

// complexes

#include <Complex.h>

inline int is_zero(const Complex& z) 
 {return is_zero(z.real())&&is_zero(z.imag());}
inline int is_approx_zero(const Complex& z)
 {return is_approx_zero(z.real())&&is_approx_zero(z.imag());}

//////////////////////////////////////////////////////////////////
#endif // reals and complexes
//////////////////////////////////////////////////////////////////


#endif
