//
// marith.cc      Integer lists and Integer overloaded arith.h
//                  using LiDIA bigints OR GNU++ Integers

#include <fstream.h>
#include "marith.h"

// Utilities for debugging output -- for example, from gdb you can give the command
//   p show(a)
// to see a where a is a bigint or bigintArray

bigint show(const bigint& a) {cout<<a<<endl; return a;}
bigintArray show(const bigintArray& a) {cout<<a<<endl; return a;}
 
#if defined(LiDIA_INTS) // then we can use the LiDIA bigint functions

bigint bezout(const bigint& aa, const bigint& bb, bigint& xx, bigint& yy) 
  {return xgcd(xx,yy,aa,bb);}
int divides(const bigint& a, const bigint& b, bigint& q, bigint& r)
  { div_rem(q,r,a,b); return is_zero(r);}
int divides(const bigint& a, long b, bigint& q, long& r)
  { div_rem(q,r,a,b); return (r==0);}

#else // we use our own

bigint bezout(const bigint& aa, const bigint& bb, bigint& xx, bigint& yy)
{bigint a,b,c,x,oldx,newx,y,oldy,newy,q;
 oldx = 1; oldy = 0; x = 0; y = 1; a = aa; b = bb;
 while (b!=0)
 { q = a/b; 
   c    = a    - q*b; a    = b; b = c;
   newx = oldx - q*x; oldx = x; x = newx;
   newy = oldy - q*y; oldy = y; y = newy;
  }
 if (a<0) {xx=-oldx; yy=-oldy; return(-a);}
 else     {xx= oldx; yy= oldy; return( a);}
}

int divides(const bigint& a, const bigint& b, bigint& q, bigint& r)
  { divide(a,b,q,r); return is_zero(r);}
int divides(const bigint& a, long b, bigint& q, long& r)
  { divide(a,b,q,r); return (r==0);}
#endif

#if defined(LiDIA_INTS)
// LiDIA bigints have no setbit function, so we add one:
void setbit(bigint& a, int e)
{
  if(e<0) return;
  if(a.bit(e)) return;
  or(a,bigint(1)<<e,a);          // replces e'th bit of a by 1
}
#endif

// oddsqrt works on odd n, called by isqrt
//
//#define oddsqrt(root,n) sqrt(root,n)   // builtin
//#define oddsqrt(root,n) sqrtq2(root,n) // 2-adic
#define oddsqrt(root,n) sqrtnr(root,n) // JC's Newton

// 2-adic version of isqrt:

int sqrtq2(bigint& root, const bigint& n)
{
  bigint a,r;
  divides(n,8,r,a);
  if(a!=1) return 0;              // odd squares must be 1 mod 8
  if(r==1) {a=3; r=0;}           // special case

//  cout<<"odd part 1 mod 8 with quotient r = " << r << endl;
  bigint twok(8), twok3(1);
  long kminus1=2;
  while(r>0)
    {
      if(even(r))
	{
	  rshift(r,1,r);
	}
      else
	{ 
	  sub(r,a,r); rshift(r,1,r); sub(r,twok3,r);
	  setbit(a,kminus1);
	}
      lshift(twok,1,twok);  lshift(twok3,1,twok3);
      kminus1++;
//      cout<<"a="<<a<<", r="<<r<<endl;
    }
  if(is_zero(r))     {root=a; return 1;}
  if(r+a==(twok>>2)) {root=(twok>>1)-a; return 1;}
  return 0;
}

// Newton-R-type iteration as in Henri Cohen's book pp 38-39

int sqrtnr(bigint& root, const bigint& n)
{
  long l = lg(n);
  bigint x,y,one(1);
  lshift(one,1+(l/2),x);   // first approx, > sqrt(n)
//  cout << "x = " << x << endl;
  while(1)
    {
      div(n,x,y); addx(y,x,y); rshift(y,1,y);
      if(y<x) {x=y;} else break;
//      cout << "x = " << x << endl;
    }
  root = x;
  return (x*x==n);
}

int isqrt(const bigint& in, bigint& root)
{
//  cout<<"In isqrt with n = " << in << endl;
  root=0;
  if(sign(in)<0) return 0;
  if(sign(in)==0) return 1;
  long twopow=0;
  bigint n(in);
  while(even(n)) {n>>=1; twopow++;}
//  cout << "2-power = " << twopow << endl;
  if(odd(twopow)) return 0;            // 2 | a to an odd power
  twopow>>=1;                     // power of 2 in root
  if(oddsqrt(root,n)) {lshift(root,twopow,root); return 1;}
  return 0;
}

#if !defined(LiDIA_INTS)
// Then we must include some functions otherwise in LiDIA
// The following code is borrowed from LiDIA file ressol.c
// simplified to work under libg++
//
// Input:
//    prime number p
//    0 <= a < p
//
// Output:
//    square root r of a mod p
//    0 <= r < p
// 
// Condition: p = 2^s * (2k+1) + 1
//	      where s fits in a long
//
// Algorithm: Shanks-Ressol
//

long jacobi(const bigint& a, const bigint& p)
{
  return legendre(a,p);
}

power_mod(bigint& ans, const bigint& base, const bigint& expo, const bigint& m)
{
  bigint b(base), e(expo);
  ans=1;
  while(e>0)
  {
    if(odd(e)) {ans*=b; ans%=m;}
    b*=b; b%=m;
    e>>=1;
  }
}

void ressol(bigint & x, const bigint & a, const bigint & p)
{
  bigint v, one(1), two(2);
   
  if ( p == two ) { x=a;  return;   }
 
// (1) p = 3 mod 4 
  if ( (p%4) == 3 ) {
    if (jacobi(a, p) == 1) {
      v = (p + one)/4;
      power_mod(x, a, v, p);
      return;
    }
    else
      {
	cerr<<"In ressol: a = " << a << " is not a quadratic residue";
	cerr<< " mod p = "<<p<<endl;
	exit(1);
      }
  }

  bigint r, n, c, z, k;
  long   s, t;

  // (2) initialisation:
  // compute k, s : p = 2^s (2k+1) + 1 
  // (MM: shift_right as part of the loop)
  k = p - one;
  s = divide_out(k,two);
  
  k = (k-one)/two;
  
  // (3) initial values
  power_mod(r, a, k, p);
  
  square(n, r);
  n %= p; // n = (r * r) % p;
  n *= a;
  n %= p; // n = (n * a) % p;
  r *= a;
  r %= p; // r = (r * a) % p;
  
  if ( n==one ) {
    x=r;
    return;
  }
  
  // (4) non-quadratic residue
  z=two;
  while ( jacobi(z,p) == 1 ) z+=one;
  
  v=(k*two)+one; // v = (k << 1) + 1;
  power_mod(c, z, v, p);
  
  // (5) iteration
  while (n > one)
    {
      k  = n;
      t  = s;
      s  = 0;
      
      while ( k!=one )
	{
	  square(k, k);
	  k%= p;         // k = (k * k) % p; 
	  s++; 
	}
      
      t -= s; 
      if (t == 0)
	{
	  cerr<<"In ressol: a = " << a << " is not a quadratic residue";
	  cerr<< " mod p = "<<p<<endl;
	  exit(1);
	}
       
      v = one<<(t-1);     // v = 1 << (t-1);
      power_mod(c, c, v, p);
      r *= c;
      r %= p;             // r = (r * c) % p;
      square(c, c);
      c %= p;             // c = (c * c) % p;
      n *= c;
      n %= p;             // n = (n * c) % p;
    } 
  
  x=r;
  return;
} 

#endif

int modsqrt(const bigint& a, const bigintArray& bplist, bigint& x)
     // Solves x^2=a mod b, returns success/fail
{
  // Assumes b square-free, primes factors in bplist
  bigint u, v, p, amodp, xmodp, m(1), two(2);
  int res=1;  x=0;
  long ip, np=bplist.length;
  for(ip=0; res&&(ip<np); ip++)
    {
      p=bplist(ip);
      if(p==two)
	{
	  xmodp=odd(a);
	}
      else // odd p
	{
	  amodp = a%p;
	  if(is_zero(amodp)) xmodp=0;
	  else
	    {
	      if(jacobi(amodp,p)==-1) return 0;
	      if(amodp<0) amodp+=p;  // since ressol wants it between 0 and p-1
	      ressol(xmodp,amodp,p);
	    }
	}
      // Now Chinese xmodp with previous (x mod m)
      bezout(m,p,u,v);
      x = x*v*p+xmodp*u*m;
      m*=p;
      x = mod(x,m);
    }
  return 1;
}

bigint roundover(const bigint& a, const bigint& b)
     // return nearest int to a/b; b must be nonzero
{
  return (a-posmod(a,b))/b;
}

bigintArray::bigintArray(long n) 
{
  items=new bigint[n]; length=n;
}

bigintArray::~bigintArray()
{
  delete[] items; length = 0; 
}

bigintArray::bigintArray(const bigintArray& l) 
{
  length=l.length;
  items=new bigint[length]; 
  for(int i=0; i<length; i++) items[i]=l.items[i];
}

bigintArray::bigintArray(const bigintArray& l, long n) 
{
  length= n<=l.length ? n : l.length;
  items=new bigint[length]; 
  for(int i=0; i<length; i++) items[i]=l.items[i];
}

bigintArray& bigintArray::operator=(const bigintArray&l)
{
  if (this==&l) return *this;
  if(length!=l.length)
    {
      delete[] items;
      length=l.length;
      items=new bigint[length]; 
    }
  for(int i=0; i<length; i++) items[i]=l.items[i];
  return *this;
}

bigint& bigintArray::get(long i) 
{
  if ((i>=0)&&(i<length)) return items[i]; 
  else{cerr<<"bad index "<<i<<" in bigintArray!\n"; return items[0];}
}

bigint bigintArray::operator()(long i) const 
{
  if ((i>=0)&&(i<length)) return items[i]; 
  else{cerr<<"bad index "<<i<<" in bigintArray!\n"; return 0;}
}

void bigintArray::set(long i, const bigint& val) 
{
  if ((i>=0)&&(i<length)) items[i]=val; 
  else {cerr<<"bad index "<<i<<" in bigintArray!\n";}
}

void bigintArray::output(ostream& os) const
{
  for(int i=0; i<length; i++) 
    {
      if(i)os<<", ";
      os<<items[i];
    }
}

ostream& operator<<(ostream& os, const bigintArray& l)
{l.output(os);return os;}


long bigintArray::locate(const bigint& num) const
{
  for ( int i=0; i<length; i++) if (num == items[i]) return i;
  return -1;
}
  
//
// bigint divisor lists etc
//

// extra_primes is an array to hold extra big primes found or added manually.  
// Its physical capacity is maxnxprimes, and currently holds nxprimes primes.

bigint* extra_primes; long nxprimes=0, maxnxprimes=0;

void add_extra_prime(const bigint& p) 
{ 
  long i;
  if(p<=BIGGESTPRIME) return;
  if(maxnxprimes==0)  // first time an extra prime has been added
    {
      maxnxprimes+=10; 
      extra_primes=new bigint[maxnxprimes];
    }
  for(i=0; i<nxprimes; i++) if(p==extra_primes[i]) return;
  if(nxprimes==maxnxprimes) // list is full, expand it
    {
      bigint* xxp=new bigint[maxnxprimes+10];
      for(i=0; i<maxnxprimes; i++) xxp[i]=extra_primes[i];
      delete[] extra_primes;
      extra_primes=xxp;
      maxnxprimes+=10;
    } 
  extra_primes[nxprimes++] = p;
  //    cerr<<"\n(Adding extra prime "<<p<<" to stored table: ";
  //    for(i=0; i<nxprimes; i++) cerr<<extra_primes[i]<<" ";
  //    cerr<<")"<<endl;
}

void show_extra_primes()
{
  cout << "Extra primes in list: ";
  long i;
  for(i=0; i<nxprimes; i++) cout << extra_primes[i] << " ";
  cout << endl;
}

void initprimes(char* pfilename, int verb)
{
  ifstream pfile(pfilename);
  if(!pfile)  // default
    {
      initprimes(1000000); 
      if(verb) 
	{
	  cout<<"Computed " << NPRIMES << " primes, ";
	  cout << "largest is " << BIGGESTPRIME << "\n";
	}
      return;
    } 
  long np;
  pfile>>np; initprimes(np);
  if(verb) 
    {
      cout<<"Computed " << NPRIMES << " primes, ";
      cout << "largest is " << BIGGESTPRIME << "\n";
    }
  bigint xp;
  while(pfile>>xp, (xp!=0) )
	{
	  add_extra_prime(xp);
	  if(pfile.eof()) break;
	}
  if(verb) show_extra_primes();
}

#if defined(LiDIA_INTS)
#include "LiDIA/rational_factorization.h"
bigintArray pdivs(const bigint& number)
{
//  cout << "In LiDIA pdivs() with number = " << number << endl;
  bigint n(number);
  if(sign(n)<0) n.negate();
  if(n<2)  // because otherwise LiDIA says that 1 has 1 as a prime factor
           // which messes things up...
    {
      bigintArray plist(0);
      return plist;
    }
  rational_factorization f(n);
//  f.verbose(1);  // for testing only
// First use any primes from the "extra" list
  long i;
  for(i=0; i<nxprimes; i++)
    {
      bigint mp=extra_primes[i];
      f.refine(mp);
    }
//  cout << "After using extra primes: " << f << endl;
  f.factor();
//  cout << "After calling factor():   " << f << endl;
  long np = f.no_of_comp();
  bigintArray plist(np);
  for(i=0; i<np; i++) plist.set(i,f.base(i));
  return plist;
}
#else
bigintArray pdivs(const bigint& number)
{
//cout<<"In pdivs() with number = " << number << endl;
  const bigint one(1);
  bigintArray plist(30);
  primevar pr;
  bigint I = abs(number), q, mp, mr;
  long np=0, i, p, r;
  while ( ( I > one ) && (pr.ok()) )
    { p = pr.value; pr++; mp=p;
      if (divides(I,p,q,r)) 
	{
	  plist.set(np++,mp);
	  I=q;
	  divide_out(I,p);
	}
// Now we might be able to conclude that the cofactor is prime:
      if(I>one) 
	if (sqr(mp)>I) 
	  {plist.set(np++,I); add_extra_prime(I); I=one; }
    }
// Now try  to find  a prime divisor from the "extra" list
  for(i=0; (i<nxprimes)&&(I>one); i++)
    {
      mp=extra_primes[i];
      if(divides(I,mp,q,mr)) 
	{
	  plist.set(np++,mp);
	  I=q;
	  divide_out(I,mp);
	}
    }
// Now we might again be able to conclude that the cofactor is prime:
  if(I>one) 
    if (sqr(bigint(BIGGESTPRIME))>I) 
      {plist.set(np++,I); add_extra_prime(I); I=one; }

  if (I>one) // failed to factor it 
    {
      cerr<<"\n***No prime factor found for "<<I<<" using trial division\n";
      cerr<<"***Proceeding with the assumption that this is prime!"<<endl;
      plist.set(np++,I);
    }
  plist.truncate(np);
//cout<< "pdivs returns " << plist << endl;
  return plist;
}

#endif

bigintArray posdivs(const bigint& number, const bigintArray& plist)
{
 int np = plist.length;
 int e, nu = 1; int nd=nu;
 int* elist = new int[np]; bigint p;
 bigintvar pr(plist);
 for(; pr.ok(); pr++)
   {p=pr.value();  elist[pr.index]=e=val(p,number); nd*=(1+e);}
 bigintArray dlist(nd);
 dlist[0]=1;
 nd=nu;
 for(pr.init(plist); pr.ok(); pr++)
   {
     p = pr.value();  e = elist[pr.index];
     for (int j=0; j<e; j++)
       for (int k=0; k<nd; k++)
         dlist[nd*(j+1)+k] = p*dlist[nd*j+k];
     nd*=(e+1);
   }
 delete[] elist;
 return dlist;
}

bigintArray alldivs(const bigint& number, const bigintArray& plist)
{
 int np = plist.length;
 int e, nu = 2; int nd=nu;
 int* elist = new int[np]; bigint p;
 bigintvar pr(plist);
 for(; pr.ok(); pr++)
   {p=pr.value(); elist[pr.index]=e=val(p,number); nd*=(1+e);}
 bigintArray dlist(nd);
 dlist[0]=1; dlist[1]=-1;
 nd=nu;
 for(pr.init(plist); pr.ok(); pr++)
   {
     p = pr.value();  e = elist[pr.index];
     for (int j=0; j<e; j++)
       for (int k=0; k<nd; k++)
         dlist[nd*(j+1)+k] = p*dlist[nd*j+k];
     nd*=(e+1);
   }
 delete[] elist;
 return dlist;
}

bigintArray sqdivs(const bigint& number, const bigintArray& plist)
{
 int np = plist.length;
 int e, nu = 1; int nd=nu;
 int* elist = new int[np]; bigint p;
 bigintvar pr(plist);
 for(; pr.ok(); pr++)
   {p=pr.value(); elist[pr.index]=e=val(p,number)/2; nd*=(1+e);}
 bigintArray dlist(nd);
 dlist[0]=1; 
 nd=nu;
 for(pr.init(plist); pr.ok(); pr++)
   {
     p = pr.value();  e = elist[pr.index];
     for (int j=0; j<e; j++)
       for (int k=0; k<nd; k++)
         dlist[nd*(j+1)+k] = p*dlist[nd*j+k];
     nd*=(e+1);
   }
 delete[] elist;
 return dlist;
}

bigintArray sqfreedivs(const bigint& number, const bigintArray& plist)
{
 int np = plist.length;
 int e, nu = 1; int nd=nu;
 int* elist = new int[np]; bigint p;
 bigintvar pr(plist);
 for(; pr.ok(); pr++)
   {elist[pr.index]=e=1; nd*=(1+e);}
 bigintArray dlist(nd);
 dlist[0]=1; 
 nd=nu;
 for(pr.init(plist); pr.ok(); pr++)
   {
     p = pr.value();  e = elist[pr.index];
     for (int j=0; j<e; j++)
       for (int k=0; k<nd; k++)
         dlist[nd*(j+1)+k] = p*dlist[nd*j+k];
     nd*=(e+1);
   }
 delete[] elist;
 return dlist;
}

void sqfdecomp(const bigint& a, bigint& a1, bigint& a2, bigintArray& plist)
     // a must be non-zero, computes square-free a1 and a2>0 such that a=a1*a2^2
     // plist will hold prime factors of a1
{
  plist = pdivs(a);
  sqfdecomp(a,plist,a1,a2);
}

void sqfdecomp(const bigint& a, bigintArray& plist, bigint& a1, bigint& a2)
     // a must be non-zero, computes square-free a1 and a2>0 such that a=a1*a2^2
     // plist already holds prime factors of a
     // plist will hold prime factors of a1
{
  long ip, j, np = plist.length, mp=0;
  bigintArray aplist(np);   // will get truncated later
  a1=1;  a2=1;
  for(ip=0; ip<np; ip++)
    {
      bigint p = plist(ip);
      long e = val(p,a);
      if(e&1) {a1*=p; aplist[mp++] =p;}
      e >>= 1;
      for(j=0; j<e; j++) a2*=p;
    }
  if(is_negative(a)) a1.negate();
  aplist.truncate(mp);
  plist=aplist;
}



//
// general functions
//
#ifndef LiDIA
#define BIG 100000   // not used in new version

bigint Ifloor(double x)  // bigfloats are just doubles in this case
{
 bigint ans(0);
 int s=sign(x); 
 if(s==0) return ans;
 if(s<0) x=-x;
/*  old version
 static double d_big = BIG;
 static bigint I_big = BIG;
 bigint shift(1); int sh=1;
 while(x>0.0)
   {
     double highx = floor(x/d_big);
     double lowx  = x-highx*d_big;
     cout << "lowx = " << lowx << endl;
     long lowa;
     if(fabs(lowx)<0.001) lowa=0; 
     else
       if(sh) 
	 {
	   if(s>0) lowa = (long)floor(lowx);
	   else    lowa = (long)ceil(lowx); 
	 }
       else
	 { 
	   lowa = (long)floor(lowx);
	 }
     ans+=lowa*shift;
     x=highx;
     shift*=I_big; sh=0;
   }
*/
// New version, well thought out and tested bit by bit:
//#define DEBUG_IFLOOR
 int e;
 double y = frexp(x,&e);
#ifdef DEBUG_IFLOOR
 cout<<"x="<<x<<", e="<<e<<endl;
#endif
 if(e>53)  // precision loss -- issue warning
   {
     cerr<<"Warning in Ifloor("<<s*x<<"): possible precision loss in "
         <<"converting to bigint; maximium rounding error "
	 << "2^"<<(e-53)<<"-1 = "<<pow(bigint(2),e-53)-1<<endl;
   }
 while((x>0)&&(e>0))   // This was >=0 in double->Integer original version
   {
#ifdef DEBUG_IFLOOR
     cout<<"x="<<x<<", e="<<e;
#endif
     setbit(ans,e-1);
#ifdef DEBUG_IFLOOR
     cout<<", ans="<<ans;
#endif
     x=fmod(x,ldexp(1,e-1)); 
#ifdef DEBUG_IFLOOR
     cout<<", new x="<<x<<endl;
#endif
     y = frexp(x,&e);
   }
 if((x>0)&&(s<0)) ++ans;        // adjust if fractional part non-zero
 if(s<0) ans.negate();  // adjust if negative
 return ans; 
}

// even older version -- don't use!

// #define BIG 100000000
// 
// bigint Ifloor(double x)
// {
//  double big=BIG;
//  bigint ans(0);
//  int s=sign(x); 
//  if(s==0) return ans;
//  if(s<0) x=-x;
//  double highx = x/big;  highx=floor(highx);
//  double lowx  = x-(highx*big);
//  long lowa = (s>0)? (long)floor(lowx): (long)ceil(lowx); 
//  long higha=(long)highx;
//  ans+=higha;  ans*=BIG; ans+=lowa;
//  if(s<0) negate(ans);
//  return ans; 
// }
#endif

#ifdef XXXXLiDIA  // never true

bigint mod(const bigint& a, const bigint& bb)
{bigint c = a%bb; 
 bigint b(bb);
 if (b<0) negate(b);
 if (c>b/2) c=c-b;
 return c;
}

long mod(const bigint& a, long b)
{long c;
 if (b<0) b=-b;
 long la = I2long(a%bigint(b));
 if (la>=0) {c=la%b;}
 else {c=(-la)%b; if (c>0) c=b-c;}
 if (c>b/2) c=c-b;
 return c;
}

#else  // always true

bigint mod(const bigint& a, const bigint& bb)
{bigint c; bigint b(bb);
 if (sign(b)<0) negate(b);
 if (sign(a)>=0) {c=a%b;}
 else {c=(-a)%b; if (sign(c)>0) c=b-c;}
 if (c+c>b) c-=b;
 return c;
}

long mod(const bigint& a, long b)
{long c;
 if (b<0) b=-b;
 long la = I2long(a%bigint(b));
 if (la>=0) {c=la%b;}
 else {c=(-la)%b; if (c>0) c=b-c;}
 if (2*c>b) c=c-b;
 return c;
}

#endif

bigint posmod(const bigint& a, const bigint& b)
{bigint c;
 if (sign(a)>=0) {c=a%b;} 
 else {c=(-a)%b;
       if (sign(c)>0) c=b-c;}
 return c;
}

long posmod(const bigint& a, long b)
{long c;
 long la = I2long(a%bigint(b));
 if (la>=0) {c=la%b;} 
 else {c=(-la)%b;
       if (c>0) c=b-c;}
 return c;
}

int divide_exact(const bigint& aa, const bigint& bb, bigint& c)
     // c = a/b with error message if remainder is non-zero
{
  bigint a(aa), b(bb), r;  // BECAUSE DIVIDE() WAS CHANGING A!!!!
  //  cout<<"In divide_exact with a = " << a << ", b = " << b << endl;
  int ok = divides(a,b,c,r);
  //  cout<<"After divides(),  a = " << a << ", b = " << b << ", q = " << c << ", r = " << r << endl;
  if(!ok)
    {
      cout<<"Error in dividing "<<a<<" by "<<b<<": not exact, remainder = "<<r<<endl;
    }
  return ok;
}

long divide_out(bigint& a, const bigint& d)
// divides a by d as many times as possible returning number of times (but none if a=0!)
{
  if(is_zero(a)) return 0;
  bigint q, r;
  long count=0;
  while(divides(a,d,q,r)) {a=q; count++;}
  return count;
}

long val(const bigint& factor, const bigint& number)
{
 bigint n = abs(number); bigint f = abs(factor);
 if (is_zero(n) || (f<bigint(2))) return 99999;  // error condition! N.B. This value 
                                     // must be unlikely and POSITIVE.
 long e = divide_out(n,f);
 return e;
}

long val(long factor, const bigint& number)
{
 bigint n = abs(number); long f = abs(factor);
 if (is_zero(n) || (f<2)) return 99999;  // error condition! N.B. This value 
                                     // must be unlikely and POSITIVE.
 int e = 0;
 while (is_zero(n%f)) {e++; n/=f; }
 return e;
}


int div(const bigint& factor, const bigint& number) 
{ if (is_zero(factor)) return is_zero(number);
  else return (is_zero(number%factor));
}

int div(long factor, const bigint& number) 
{ if (factor==0) return is_zero(number);
  else return is_zero(number%factor);
}


long bezout(const bigint& aa, long bb, bigint& xx, bigint& yy)
{bigint a,b,c,x,oldx,newx,y,oldy,newy,q;
 oldx = 1; oldy = 0; x = 0; y = 1; a = aa; b = bb;
 while (sign(b)!=0)
 { q = a/b; 
   c    = a    - q*b; a    = b; b = c;
   newx = oldx - q*x; oldx = x; x = newx;
   newy = oldy - q*y; oldy = y; y = newy;
  }
 if (sign(a)<0) {xx=-oldx; yy=-oldy; return -I2long(a);}
 else     {xx= oldx; yy= oldy; return  I2long(a);}
}

 
bigint invmod(const bigint& a, const bigint& p)
{bigint g,x=0,y=0,one=1;
 g=bezout(a,p,x,y);
 if (g==one) return x;
 else
   {cerr << "invmod called with " << a << " and " << p << " -- not coprime!\n";
    return 0; }
}

long invmod(const bigint& a, long p)
{bigint g,x=0,y=0,one=1;
 g=bezout(a,p,x,y);
 if (g==one) return I2long(x);
 else
   {cerr << "invmod called with " << a << " and " << p << " -- not coprime!\n";
    return 0; }
}


int m1pow(const bigint& a)
{ return (odd(a) ?  -1 : +1);
}
 
static int table8[8] = {0,1,0,-1,0,-1,0,1};
static int table4[4] = {0,1,0,-1};
static int table44[4][4] = {{0,0,0,0}, {0,1,0,1}, {0,0,0,0}, {0,1,0,-1}};
  
int chi2(const bigint& a)
{ 
  return table8[posmod(a,8)];
}
 
int chi4(const bigint& a)
{ 
  return table4[posmod(a,4)];
} 


int hilbert(const bigint& a, const bigint& b)
{ 
  return table44[posmod(a,4)][posmod(b,4)];
}

int hilbert(const bigint& a, long b)
{ 
  return table44[posmod(a,4)][posmod(b,4)];
}

int hilbert(long a, const bigint& b)
{ 
  return table44[posmod(a,4)][posmod(b,4)];
}


static int leg(const bigint& a, const bigint& b) 
//nb this function is not intended for public use!
{ bigint aa = a;
  bigint bb = b;
//  cout<<"leg("<<a<<","<<b<<") = "<<flush;
  bigint c, one(1);
  int ans = 1;
  while (bb>one) 
  {     aa = aa % bb;
        if (sign(aa)<0)       {aa=-aa; ans*=chi4(bb);}
        while (is_zero(aa%4)) {aa/=4;}
        if (is_zero(aa%2))    {aa/=2; ans *= chi2(bb);}
        ans*=hilbert(aa,bb);
        c=bb; bb=aa; aa=c;
  }
//  cout << ans << endl;
  return ans;
}
 

int legendre(const bigint& a, const bigint& b)
{ 
  return ((is_one(gcd(a,b)) && (odd(b))) ? leg(a,b) : 0);
}

int legendre(const bigint& a, long b)
{ 
  return ((is_one(gcd(a,(bigint)b)) && (b%2)) ? leg(a,b) : 0);
}

 
int kronecker(const bigint& d, const bigint& n)
{ int ans=0; bigint m=n; long d4=posmod(d,4);
  if (is_one(gcd(d,n)) && ((d4==0)||(d4==1)))
  { ans=1;
    while (is_zero(m%4)) m/=4;
    if (is_zero(m%2)) {m/=2; ans*=(posmod(d,8)==1 ? 1 : -1);}
    ans *= legendre(d,m);
  }
  return ans;
}
 
int kronecker(const bigint& d, long n)
{ int ans=0; bigint m=n; long d4=posmod(d,4);
  if ((gcd(d,n)==1) && ((d4==0)||(d4==1)))
  { ans=1;
    while (is_zero(m%4)) m/=4;
    if (is_zero(m%2)) {m/=2; ans*=(posmod(d,8)==1 ? 1 : -1);}
    ans *= legendre(d,m);
  }
  return ans;
}

long gcd(const bigint& a, long b)
{
  bigint bb=b; return I2long(gcd( a, bb )); 
}

int modrat(const bigint& n, const bigint& m, const bigint& lim, 
           /* return values: */ bigint& a, bigint& b)
{
 bigint q,r,t,qq,rr,tt,quot; int sucess; // cout<<"modrat("<<n<<","<<m<<")";
 q=m; r=posmod(n,m); qq=0; rr=1; t=0; tt=0; a=r; b=1; 
 if (r<lim) 
   { 
//   cout<<" = "<<a<<"/"<<b<<"\n";
     return 1;
   }
 while (sign(r)!=0) 
 { 
   divides(q,r,quot,t);
   q = r;   r = t;
   tt = qq-quot*rr; qq = rr; rr = tt;
   if (r<lim)
     {
       if (abs(rr)<lim) {a=r; b=rr; return 1;}
       else cerr << "modrat error: no reconstruction for " << n << " mod " << m << "\n";
//     cout<<" = "<<a<<"/"<<b<<"\n";
       return 0;
     }
 }
 cerr << "modrat error: common factor with " << n << " mod " << m << "\n";
 return 0;
}


// bigcomplex-related function:

bigintArray Introotscubic(const bigint& a, const bigint& b, const bigint& c)
{ bigcomplex za(I2bigfloat(a)), zb(I2bigfloat(b)), zc(I2bigfloat(c));
  bigcomplex* croots =  solvecubic(za,zb,zc);
  bigintArray iroots(3);
  int niroots =0; bigint x,cx;
  for (int i=0; i<3; i++)
    {x = Iround(real(croots[i]));
    if (is_zero(x)) 
    {
      if (is_zero(c)) 
        {
         int newone=1;
         for(int j=0; (j<niroots)&&newone; j++) newone=(x!=iroots[j]);
         if(newone) iroots[niroots++]=x;
        }
    } 
    else
        {cx = c/x;
         if (x*cx==c)
            if ( is_zero((x+a)*x+b+cx) )
              {//x is a root, so add to list if not there already:
               int newone=1;
               for(int j=0; (j<niroots)&&newone; j++) newone=(x!=iroots[j]);
               if(newone) iroots[niroots++]=x;
              }
        }
     }
  delete[] croots;
  iroots.truncate(niroots);
  return iroots;  
}

bigintArray Introotsquartic(const bigint& a, const bigint& b, const bigint& c, const bigint& d)
{
  bigfloat xb(I2bigfloat(a)), xc(I2bigfloat(b)), xd(I2bigfloat(c)), xe(I2bigfloat(d));
  bigcomplex* croots =  solverealquartic(bigfloat(1),xb,xc,xd,xe);
  bigintArray iroots(4);
  int niroots =0; bigint x,dx;
  for (int i=0; i<4; i++)
    {
      bigcomplex z = croots[i];
#ifdef DEBUG
      cout << "Complex root number " << (i+1) << " = " << z << "\n";
#endif      
      if(!is_real(z))
	{
#ifdef DEBUG
	  cout << "--not real\n ";
#endif      
	  continue;
	}
      else
	{
	  x = Iround(real(z));
#ifdef DEBUG
	  cout << "rounds to " << x << "\n";
#endif      
	  if (is_zero(x)) 
	    {
	      if (is_zero(d)) 
		{
		  int newone=1;
		  for(int j=0; (j<niroots)&&newone; j++) newone=(x!=iroots[j]);
		  if(newone) iroots[niroots++]=x;
		}
	    } 
	  else
	    {dx = d/x;
	     if (x*dx==d)
	       if ( is_zero(((x+a)*x+b)*x+c+dx) )
		 {//x is a root, so add to list if not there already:
		   int newone=1;
		   for(int j=0; (j<niroots)&&newone; j++) newone=(x!=iroots[j]);
		   if(newone) iroots[niroots++]=x;
		 }
	   }
	}
    }
  delete[] croots;
  iroots.truncate(niroots);
  return iroots;  
}

#ifdef 0  // 0bsolete stuff:

bigint primdiv(const bigint& number) //first prime divisor (else self)
{
  bigint a = abs(number);
  primevar pr;
  bigint mp, q, mr;
  long p,r;
  while ( pr.ok()  )
    {p = pr; pr++; mp=p;
     if(divides(a,p,q,r)) {return mp;}
     if ( mp*mp > a ) {add_extra_prime(a);return a;}
    }
// Now try  to find  a prime divisor from the "extra" list
  long i;
  for(i=0; i<nxprimes; i++)
    {
      mp=extra_primes[i];

      if(divides(a,mp,q,mr)) {return mp;}
    }
  cerr<<"\n***No prime factor found for "<<number<<"\n";
  cerr<<"***Proceeding with the assumption that this is prime!"<<endl;
  return a;
}

// exceptional case finds long primes of bigint argument

longlist pdivs_to_long(const bigint& number)
{
  const bigint one(1);
  longlist plist(30);   // surely this is big enough !
  int np = 0;
  primevar pr;
  bigint I = abs(number), q, mp;
  bigint rootnum = sqrt(I);  // rounded down
  long p, r;
  while ( ( I > one ) && (pr.ok()) )
    { p = pr; pr++; mp=p;
      if (divides(I,p,q,r)) 
	{
	  plist.set(np++, p);
	  I=q;
	  divide_out(I,p);
	}
      if (I>one) 
	if (mp>rootnum) 
	  {plist.set(np++, I2long(I)); I=one; }
    }
  if (I>one)
    {if ( is_long(I) ) plist.set(np++,I2long(I));
     else {cout << "Cannot build a list of long factors of bigint " 
                << number << endl;
           exit(1);
          }
    }  // ie in case of p-factors outside range, assume
       //the cofactor is prime.
 plist.truncate(np);
 return plist;
}

longlist pdivs_to_long(const long& ell)  // just to save silly error
{
  return pdivs(ell);
}

#endif

// end of file marith.cc
