//
// marith.h           Integer additions to arith.h
//                      Using LiDIA bigints if LiDIA is defined
//                      (in interface.h) else GNU++ Integers

// N.B. Since in LiDIA, Integer is itself a type, compatible programs
// must use bigint and NOT Integer as the type-name for multi-precision
// integers.  This is typedef'd to Integer when NOT using LiDIA.
// The LiDIA switch is set in interface.h, not here

#ifndef _MARITH_H
#define _MARITH_H

#include <stream.h>
#include "compproc.h" // which includes interface.h & LiDIA stuff
#include "arith.h"

bigint bezout(const bigint& aa, const bigint& bb, bigint& xx, bigint& yy);
int divides(const bigint& a, const bigint& b, bigint& q, bigint& r);
int divides(const bigint& a, long b, bigint& q, long& r);
// returns 1 iff remainder r==0

#if defined(LiDIA) // then we use LiDIA's rounding functions
inline bigint Iround(bigfloat x) {bigint a;        x.bigintify(a); return a;}
inline bigint Ifloor(bigfloat x) {bigint a; floor(x).bigintify(a); return a;}
inline bigint Iceil(bigfloat x)  {bigint a;  ceil(x).bigintify(a); return a;}
#else // we use our own
bigint Ifloor(double x);
inline bigint Iceil(double x)  {return -Ifloor(-x);}
inline bigint Iround(double x) {return (x>0?Ifloor(x+0.5):Iceil(x-0.5));}
#endif

#if !defined(LiDIA_INTS)
power_mod(bigint& ans, const bigint& base, const bigint& expo, const bigint& m);
void ressol(bigint & x, const bigint & a, const bigint & p);
#endif

bigint posmod(const bigint& a, const bigint& b);  // a mod b in range 0--(b-1)
long posmod(const bigint& a, long b);
int isqrt(const bigint& a, bigint& root);
int divide_exact(const bigint& aa, const bigint& bb, bigint& c);
     // c = a/b with error message if remainder is non-zero
long divide_out(bigint& a, const bigint& d);
// divides a by d as many times as possible returning number of times (but none if a=0!)

//
// bigintArrays and bigintvars
//

class bigintArray {
friend class bigintvar;
public:
        long length;        
private:
        bigint *items;  
public:
        bigintArray(long n=0);
        ~bigintArray();
        bigintArray(const bigintArray& l);
        bigintArray(const bigintArray& l, long n);
        bigintArray& operator=(const bigintArray&l);
        bigint& get(long i);
        long locate(const bigint& num) const; //return index if in list,else -1
        bigint& operator[](long i) {return this->get(i);}
        bigint operator()(long i) const;
        void set(long i, const bigint& val);
        void truncate(long newlength) {if (newlength<length) length=newlength;}
         // no need to deallocate memory here because copy constructors do so
         // almost always it is the temporary of a function that calls truncate
        void output(ostream& os) const;
        friend ostream& operator<<(ostream& os, const bigintArray& l);
};

class bigintvar {
public:
        long index;        /* current index */
private:
        bigint* values;
        long maxindex;     /* max index */
public:
        bigintvar(const bigintArray& l) 
          {maxindex=l.length-1; index=0; values=l.items;}
        void init(const bigintArray& l) 
          {maxindex=l.length-1; index=0; values=l.items;}
        int next() {if ((index++)<maxindex) {values++; return 1;}
                    else return 0;
                   }
//NB postfix and prefix versions of ++ are the same here 
//for backwards compatibility.
        bigint operator++(int)  // postfix
        {if ((index++)<=maxindex) {bigint& ans=*values; values++; return ans;}
         else return 0;
        }
        bigint operator++()    // prefix
        {if ((index++)<=maxindex) {bigint& ans=*values;values++; return ans;}
         else return 0;
        }
        int ok() const {return index<=maxindex;}
        int more() const {return index<maxindex;}
        bigint& value() const {return *values;}
//        operator bigint&() const {return *values;}
};

// Utilities for debugging output -- for example, from gdb you can give the command
//   p show(a)
// to see a where a is a bigint or bigintArray

bigint show(const bigint& a);
bigintArray show(const bigintArray& a);

//
// divisor lists  --  bigint overloads
//

extern bigint* extra_primes;  extern long nxprimes, maxnxprimes;
void add_extra_prime(const bigint& p);
void show_extra_primes();
void initprimes(char* pfilename, int verb=0);

/* obsolete functions:
bigint primdiv(const bigint& number); //first prime divisor (else self)
longlist pdivs_to_long(const bigint& number);
longlist pdivs_to_long(const long& ell);  // just to save silly error
*/

// The following uses LiDIA factorization ifdef LiDIAINTS:
bigintArray pdivs(const bigint& number);

bigintArray posdivs(const bigint& number, const bigintArray& plist);
inline bigintArray posdivs(const bigint& number)
{
 const bigintArray& plist=pdivs(number);
 return posdivs(number, plist);
}

bigintArray alldivs(const bigint& number, const bigintArray& plist);
inline bigintArray alldivs(const bigint& number)
{
 const bigintArray& plist=pdivs(number);
 return alldivs(number, plist);
}

bigintArray sqdivs(const bigint& number, const bigintArray& plist);
inline bigintArray sqdivs(const bigint& number)
{
 const bigintArray& plist=pdivs(number);
 return sqdivs(number, plist);
}

bigintArray sqfreedivs(const bigint& number, const bigintArray& plist);
inline bigintArray sqfreedivs(const bigint& number)
{
 const bigintArray& plist=pdivs(number);
 return sqfreedivs(number, plist);
}

void sqfdecomp(const bigint& a, bigint& a1, bigint& a2, bigintArray& plist);
     // a must be non-zero, computes square-free a1 and a2>0 such that a=a1*a2^2
     // plist will hold prime factors of a

void sqfdecomp(const bigint& a, bigintArray& plist, bigint& a1, bigint& a2);
     // a must be non-zero, computes square-free a1 and a2>0 such that a=a1*a2^2
     // plist already holds prime factors of a


//
// general purpose routines -- bigint overloads
//

bigint mod(const bigint& a, const bigint& b);     // a mod b in range +- half b
long mod(const bigint& a, long b);

long val(const bigint& factor, const bigint& number);
long val(long factor, const bigint& number);

int div(const bigint& factor, const bigint& number);
int div(long factor, const bigint& number);
inline int ndiv(const bigint& factor, const bigint& number)
  { return !div(factor, number);}
inline int ndiv(long factor, const bigint& number)
  { return !div(factor, number);}

long bezout(const bigint& aa, long bb, bigint& xx, bigint& yy);
bigint invmod(const bigint& a, const bigint& p);  // -a mod p
long invmod(const bigint& a, long p);

int m1pow(const bigint& a);
int chi2(const bigint& a);
int chi4(const bigint& a);
int hilbert(const bigint& a, const bigint& b);
int hilbert(const bigint& a, long b);
int hilbert(long a, const bigint& b);
int legendre(const bigint& a, const bigint& b);
int legendre(const bigint& a, long b);
int kronecker(const bigint& d, const bigint& n);
int kronecker(const bigint& d, long n);

long gcd(const bigint& a, long b);

int modrat(const bigint& n, const bigint& m, const bigint& lim, 
           /* return values: */ bigint& a, bigint& b);

int modsqrt(const bigint& a, const bigintArray& bplist, bigint& x);
     // Solves x^2=a mod b, returns success/fail

bigint roundover(const bigint& a, const bigint& b);
     // return nearest int to a/b; b must be nonzero

// bigcomplex-related functions:

bigintArray Introotscubic(const bigint& a, const bigint& b, const bigint& c);
bigintArray Introotsquartic(const bigint& a, const bigint& b, const bigint& c,
                            const bigint& d);

#endif
// end of file marith.h
