//
// points.cc, adapted from elliptic.cc by OM
//

#include <minmax.h>
#include "points.h"  // which includes curve.h
#include "cperiods.h"

//
// Point member functions
//

Point::~Point(void) {;}

void Point::reduce(void)
{
  if(sign(Z)==0){  // point at infinity
    X = 0 ;
    Y = 1 ;
    return ;
  }
  if(Z==1){  // integral point, no work needed
    return ;
  }
  bigint d = gcd(X, Y) ; d = gcd(d, Z) ; // now d = gcd(X, Y, Z)
  if(sign(d)==0){
    cerr << "##Point::reduce:  Bad point: gcd == 0 \n" ;
    X = 0 ; Y = 1 ; Z = 0;
    exit(1) ;
    return ;
  }
  if(d!=1) {
    X /=  d ;
    Y /=  d ;
    Z /=  d ;
  }
  if(sign(Z)<0){
    X.negate() ;
    Y.negate() ;
    Z.negate() ;
  }
}

// test of equality of points
int eq(const Point&P, const Point&Q) // assumes same curve
{
  bigint temp = P.X*Q.Z-P.Z*Q.X;
  int ans = (sign(temp)==0);
  if(!ans) return 0;
  temp = P.Y*Q.Z-P.Z*Q.Y;
  ans = (sign(temp)==0); 
  if(!ans) return 0;
  temp = P.Y*Q.X-P.X*Q.Y;
  ans = (sign(temp)==0); 
  return ans;
}

int Point::operator==(const Point& Q) const
{
  if(E != Q.E) return 0 ;      // different curves!
  return eq(*this,Q);
}

int Point::operator!=(const Point& Q) const
{
  return !(*this == Q) ;
}

// Shifting coords: 
// N.B. (1) This changes the curve too
//      (2) This uses coords (x/z^2,y/z^3)

void raw_shift_point(const bigint& x, const bigint& y, const bigint& z, 
                const bigint& u, const bigint& r, 
                const bigint& s, const bigint& t, 
                bigint& newx, bigint& newy, bigint& newz, int back)
{
  //cout<<"x="<<x<<", y="<<y<<", z="<<z<<endl;
  //cout<<"\t[u,r,s,t] = ["<<u<<","<<r<<","<<s<<","<<t<<"]\n";
  const bigint& z2 = z*z;
  const bigint& z3 = z*z2;
  if(back)
    {
      const bigint& u2 = u*u;
      const bigint& u3 = u*u2;
      newx = u2*x + r*z2;
      newy = u3*y + u2*s*x*z + t*z3;
      newz = z;
    }
  else
    {
      newx = x - r*z2;
      newy = y - s*x*z + (r*s-t)*z3;
      newz = u*z;
    }
  //cout<<"newx="<<newx<<", newy="<<newy<<", newz="<<newz<<endl;

  const bigint& newz2 = newz*newz;
  const bigint& v2 = gcd(newx, newz2);
  const bigint& v3 = gcd(newy, newz*newz2);
  if(sign(v2)!=0) {newx/=v2;   newy/=v3;   newz/=(v3/v2);}
  //cout<<"newx="<<newx<<", newy="<<newy<<", newz="<<newz<<endl;
}

// Shifts P via T = (u, r, s, t) on to newc.   
// N.B. Assumes that (1) T(P.E) = newc
//                   (2) T(P) is on newc
// "back" means use reverse transform

Point shift(const Point& P, Curvedata* newc, 
	    const bigint& u, const bigint& r, const bigint& s, 
	    const bigint& t, int back)
{
  if(P.iszero()) return Point(newc);
  if(!P.isvalid())
    cout << "Attempting to shift the point " << P 
	 << "which is not a valid point on its curve " << P.getcurve() << "!\n";
  const bigint& oldz = gcd(P.X, P.Z);
  const bigint& oldx = P.X/oldz;
  const bigint& oldy = P.Y;
  bigint newx, newy, newz;
  raw_shift_point(oldx, oldy, oldz, u, r, s, t, newx, newy, newz, back);
  
  Point ans(newc, newx*newz, newy, pow(newz,3));
  if(!ans.isvalid())
    {
      cout << "Result of shifting the point " << P;
      cout << " is " << ans << " which is not a valid point on its curve " 
	   << ans.getcurve() << "!\n";
    }
  return ans;
}



// addition and subtraction
void Point::operator+=(const Point& Q)  // P += Q; 
{
  Point sum = (*this) + Q ;
  E = sum.E; X = sum.X; Y = sum.Y; Z = sum.Z; reduce();
  ord = 0; height = -1;
}

void Point::operator-=(const Point& Q)  // P -= Q ; 
{
  Point sum = -Q;  sum+=(*this);
  E = sum.E; X = sum.X; Y = sum.Y; Z = sum.Z; reduce();
  ord = 0; height = -1;
}
        
Point Point::operator+(const Point& Q) const // P + Q
{
  Point ans(E);
  // make sure that both points are on the same curve
  if(E != Q.E){
    cerr << "## Can't add points on different curves!" << "\n" ;
    return ans;
  }
  // NOTE: we don't do any reductions here, but rely on assignment to do it
  // check special cases first
  if(Z==0) return Q ;
  if(Q.iszero()) return *this ;
  if(eq(*this , Q))      {return Q.twice();}
  Point minusQ=-Q;
  if(eq(*this , minusQ)) {return ans;}       // zero
        
  // we now have genuine work to do
  // let's set up some local variables to avoid repeated references
  // coefficients
  bigint A1,A2,A3,A4,A6;  E->getai(A1,A2,A3,A4,A6);
  // coordinates of P
  const bigint& X1 = X ;
  const bigint& Y1 = Y ;
  const bigint& Z1 = Z ;
  // coordinates of Q
  const bigint& X2 = Q.X ;
  const bigint& Y2 = Q.Y ;
  const bigint& Z2 = Q.Z ;
  const bigint& Z12 = Z1 * Z2 ;
        
  const bigint& L = - Y2 * Z1 + Y1 * Z2 ;    /* lambda */
  const bigint& M = - X2 * Z1 + X1 * Z2 ;    /* mu    */
  const bigint& N = - Y1 * X2 + Y2 * X1 ;    /* nu   */
        
  const bigint& Mz =  M * M * Z12  ; 
  
  const bigint& t = L * L * Z12 + M * ( A1 * L * Z12 - M * (A2 * Z12
                                        + X1 * Z2 + X2 * Z1 ) )  ;
  const bigint& newX =  M * t ;
  const bigint& newY = - (  t * (L + A1 * M) +   Mz * (N + A3 * M )) ;    
  const bigint& newZ = M * Mz ;
  ans.init(E, newX, newY, newZ);
  return ans;
}

Point Point::operator-(const Point& Q) const // P - Q
{
  Point ans = (-Q) ;
  ans += (*this);
  return ans;
}

Point Point::twice(void) const // doubles P
{
  // do trivial cases
  Point ans(E);
  if( Z==0) return ans;
  bigint A1,A2,A3,A4,A6;  E->getai(A1,A2,A3,A4,A6);
  Point minusthis = -(*this);
  if(eq(*this,minusthis)) return ans;   // order 2

  const bigint& Zsq = Z * Z ;
  
  const bigint& L = 3 * X * X  +  2 * A2 * X * Z  +  
                                     A4 * Zsq  -  A1 * Y * Z ;
  const bigint& M = 2 * Y  +  A1 * X  +  A3 * Z ;
        
  const bigint& Mz = M * Z ;
  const bigint& N =  - X * X * X -A3 * Y * Zsq  + A4 * X *Zsq  +  
    2 * A6 * Z *Zsq ;
        
  const bigint& t  = L * L  +  Mz * ( A1 * L  -  M * ( A2 * Z  + 2 * X ) );  

  const bigint& newX = t * Mz ;
  const bigint& newY = - ( L * t +  Mz * ( A1 * t+ M * ( N  + A3 * Mz * Z) ) );
  const bigint& newZ = Mz * Mz * Mz;

  ans.init(E, newX, newY, newZ) ;
  return ans;
}

Point Point::operator-(void) const // -Q
{
  Point negation(*this);
  negation.Y = -Y - (E->a3)*Z - (E->a1)*X;
  return negation;
}

//inline int odd(const int n) { return n & 1 ; }

// calculates nP
Point operator*(int n, const Point& P)
{
  Point ans(*(P.E));
  if(P.iszero() || n == 0) return ans;
  int negate = (n < 0) ;
  if(negate) n = - n ;
  if(n == 1) {
    ans = P;
    if(negate) ans =  -P;
    return ans ;
  }
  // now n >= 2
  if(n == 2){
    ans = P.twice() ;
    if(negate) ans =  -ans;
    return ans ;
  }
  // now n >= 3
  if(n&1) ans = P ;    // else ans is ZERO
  Point  Q = P ; 
  while(n > 1){
    Q = Q.twice() ; // 2^k P
    n /= 2 ;
    if(n&1) ans = ans + Q ;
  }
  if(negate) ans = -ans ;
  return ans ;
}

int order(Point& p)
{
  // ASSUME that point is valid; check before calling if unknown
  if (p.ord) {return p.ord;}
  bigint eight(8), z=p.Z;
  if (sign(z)==0)  {p.ord = 1; return 1; }
  if (z>eight)     {p.ord =-1; return -1;}
  Point q = p;  long ord=1;
  while ( (sign(z)!=0) && (z<=eight) ) // (worst denom of a torsion point)
    {ord++;   q+=p;  z = q.Z;}
  if (sign(z)!=0) ord = -1;
  p.ord = ord;
  return ord;
}

int order(Point& p, PointArray& multiples)
{
  //  cout<<"In order() with p = "<<p<<endl;
  // ASSUME that point is valid; check before calling if unknown
  multiples = PointArray(12);  // = max order of a torsion point.  If a 
                               // point ever has more than 12 integral 
                               // multiples, this would give problems, so we 
                               // set a limit in the loop below.
  multiples[0] = Point(*(p.E));
  if (sign(p.Z)==0) {p.ord=1; return 1; }
  multiples[1] = p;
  int ord = 2;  Point q = p;
  bigint eight(8);
  while ( (sign(q.Z)!=0) && (q.Z<=eight) && (ord<13) )
  { 
    q+=p;
    if (sign(q.Z)!=0) multiples[ord++] = q;
  }
  multiples.truncate(ord);
  if (sign(q.Z)!=0) ord = -1;
  p.ord = ord;
  //  cout<<"Returning ord="<<ord<<", multiples = "<<multiples<<endl;
  return ord;
}

int Point::isvalid() const // P on its curve ?
{
  if(E == 0){
    cerr << "## Bad point: null curve pointer!\n" ;
    return 0 ;
  }
  if((sign(X)==0)&&(sign(Y)==0)&&(sign(Z)==0)) return 0;   // Null point, useful for terminating input
  if(sign(Z)==0) return 1 ;                    // Point at infinity is on a curve
  else{
    // should calculate 
    //                       y^2 +a1 x y + a3 y 
    //  and
    //                       x^3 + a2 x^2 + a4 x + a6
    // where 
    //          x = X/Z, y = Y/Z
    // and verify equality.
    //
    // In homogeneous coordinates:
    //
    // Lhs = Y^2 Z + a1 XYZ + a3 YZ^2 = (YZ) *(Y + a1 X + a3 Z)
    //
    //
    // Rhs = X^3 +a2 X^2 Z + a4 X Z^2 + a6 Z^3
    //
    bigint A1,A2,A3,A4,A6;    E->getai(A1,A2,A3,A4,A6);
    const bigint& Lhs = Y*Z*(Y + A1*X + A3*Z) ;
    const bigint& Rhs = A6*pow(Z,3) + X*(A4*Z*Z + X*(A2*Z + X)) ;
    return Lhs == Rhs ;
  }
}
                
PointArray::PointArray(long n) 
{
#ifdef _TRACE_
  cout<<"PointArray initializer constructor with n = "<<n<<endl;
#endif
  length=n; if (length) items=new Point[n];
}
  
PointArray::~PointArray() 
{
#ifdef _TRACE_
  cout<<"PointArray destructor called, value of items is "<<items<<endl;
#endif
  if (length) delete[] items; length = 0; 
}

PointArray::PointArray(const PointArray& l) 
{
#ifdef _TRACE_
  cout<<"PointArray copy constructor called"<<endl;
#endif
  length=l.length;
  if (length) items=new Point[length]; 
  long n=length; Point* x=items, *y=l.items;
  while(n--) {*x++=*y++;}
}

PointArray::PointArray(const PointArray& l, long n) 
{
#ifdef _TRACE_
  cout<<"PointArray partial (two-arg) copy constructor called"<<endl;
#endif
  length= n<=l.length ? n : l.length;
  if (length) items=new Point[length]; 
  n=length; Point* x=items, *y=l.items;
  while(n--) {*x++=*y++;}
}

PointArray& PointArray::operator=(const PointArray&l)
{
#ifdef _TRACE_
  cout<<"PointArray op= called, rhs pointer is "<<l.items<<endl;
#endif
  if (this==&l) return *this;
  if (length) delete[] items;
  length=l.length;
  if (length) items=new Point[length]; 
  long n=length; Point* x=items, *y=l.items;
  while(n--) {*x++=*y++;}
  return *this;
}

void PointArray::set(long i, const Point& val) 
{
  if ((i>=0)&&(i<length)) items[i]=val; 
  else {cerr<<"bad index "<<i<<" in PointArray!\n";}
}

void PointArray::truncate(long newlength) 
{ 
  if (newlength<length) length = newlength;
}
// no need to deallocate memory here because copy constructors do so
// almost always it is the temporary of a function that calls truncate

void PointArray::output(ostream& os) const
{
  long i;Point *it;
  for(i=0,it=items;i<length;i++,it++) {if(i)os<<", ";os<<(*it);} 
}

  

long PointArray::locate(const Point& p)
{
  for (int i=0; i<length; i++ ) if (p == items[i]) return i;
  return -1;
}


// find all the torsion points on a curve (Curvedata member function)

PointArray old_torsion_points(Curvedata& E);  // code is below

PointArray* Curvedata::make_torsion()
{
  PointArray *ans = new PointArray;
  *ans = torsion_points(*this);
  ntorsion = ans->getlength();
  return ans;
}

//#define DEBUG_TORSION

// Cperiods is a class containing a basis for the period lattice L;
// it knows how to compute points from z mod L; so this function
// effectively does the same as PARI's pointell() (called ellzto point
// in the next PARI release)
//
Point make_tor_pt(Curvedata& E, Cperiods& per, 
		  const bigfloat& ra1, const bigfloat& ra2, const bigfloat& ra3, 
		  const bigcomplex& z)
{
  bigcomplex cx,cy;
#ifdef DEBUG_TORSION
  cout<<"In make_tor_pt() with z = " << z << endl;
#endif
  per.XY_coords(cx,cy,z);
  cx = cx-(ra1*ra1+4*ra2)/12;
  cy = (cy - ra1*cx - ra3)/2;
#ifdef DEBUG_TORSION
  cout<<"(x,y) = ("<<(cx)<<","<<(cy)<<")\n";
#endif
  bigint x=Iround(real(cx)), y=Iround(real(cy));
  Point P(E, x, y);
  return P;
}

// New torsion point routine using Mazur's theorem to limit possibilities 
// and computing possible real torsion from the period lattice.  Suggestion 
// of Darrin Doud.
//
PointArray torsion_points(Curvedata& E)  // After Darrin Doud as adapted by JC
{
  if ( E.isnull() ) return 0;
//
// table[i][] contains a list of possible maximal orders for a point,
// given that the 2-torsion subgroup has order i
//
  static long table[5][5] = {{},{5,7,9,3},{12,6,8,4,10},{},{8,6,4}};
  static long nt[5] = {0,4,5,0,3};
#ifdef DEBUG_TORSION
  cout<<"\nIn torsion_points() with curve "<<E<<"\n";
#endif
  bigint a1,a2,a3,a4,a6,sa2,sa4,sa6, d, x, y ;
  E.getai(a1,a2,a3,a4,a6);
  bigfloat ra1=I2bigfloat(a1), ra2=I2bigfloat(a2), ra3=I2bigfloat(a3);
  long i,j,nroots,ntp,nt2;  int scaled_flag;
  PointArray points(16);    // upper bound
  PointArray two_torsion(4);
  PointArray cycle;
  Cperiods per(E);  bigcomplex w1,w2,z,z2; 
  per.getwi(w1,w2); z2=w2/2;
  int ncc = getconncomp(E);
  int found;
#ifdef DEBUG_TORSION
  cout<<"Periods: "<<per<<"\nReal Period = "<<w1<<endl;
  cout<<ncc<<" real component(s)"<<endl;
#endif
  per.norm_region();

  ntp = 1; nt2 = 1;
  points[0] = Point(E) ;     // zero point
  two_torsion[0] = Point(E) ;     // zero point
  if ( (sign(a1)==0) && (sign(a3)==0) )
    {sa2=a2; sa4=a4; sa6=a6; scaled_flag=0; }
  else
    {sa2 = a1*a1 + 4*a2;
     sa4 = 8*a1*a3 + 16*a4;
     sa6 = 16*a3*a3 + 64*a6;
     scaled_flag=1; }
  d = sa2*sa2*(sa4*sa4 - 4*sa2*sa6) + 18*sa2*sa4*sa6
                    - 4*sa4*sa4*sa4 - 27*sa6*sa6;
  Point p, q;

// First test y=0 for points of order 2:

  bigintArray xlist = Introotscubic(sa2, sa4, sa6);
  nroots = xlist.length;
  for (i=0; i<nroots; i++)
    {
      x = xlist[i];
      if (scaled_flag) p.init(E, 2*x, - a1*x - 4*a3, 8);
      else p.init(E, x, 0);
      two_torsion[nt2++] = p;
    }
  two_torsion.truncate(nt2);
#ifdef DEBUG_TORSION
  cout<<"Number of 2-torsion points = " << nt2 << endl;
  for(i=0; i<nt2; i++) cout<<two_torsion[i]<<"\t";
  cout<<endl;
#endif

  for(i=0, found=0; (i<nt[nt2])&&(!found); i++)
    {
      long ni=table[nt2][i]; z=w1/ni;
#ifdef DEBUG_TORSION
      cout<<"Looking for a point of order " << ni << "\n";
#endif
      p = make_tor_pt(E,per,ra1,ra2,ra3,z);
#ifdef DEBUG_TORSION
      cout<<"p = " << p <<"?\n";
#endif
      found=(p.isvalid())&&(order(p,cycle)==ni);
      if(!found&&(ncc==2)&&even(ni))
	{
	  p = make_tor_pt(E,per,ra1,ra2,ra3,z+z2);
#ifdef DEBUG_TORSION
	  cout<<"p = " << p <<"?\n";
#endif
	  found=(p.isvalid())&&(order(p,cycle)==ni);
	  if(!found&&((ni%4)==2))
	    {
	      p = make_tor_pt(E,per,ra1,ra2,ra3,2*z+z2);
#ifdef DEBUG_TORSION
	      cout<<"p = " << p <<"?\n";
#endif
	      found=(p.isvalid())&&(order(p,cycle)==ni);
	    }
	}
      if(found)
	{
	  ntp=ni;
	  for(j=1; j<ni; j++) points[j]=cycle[j];
#ifdef DEBUG_TORSION
	  cout<<"Found a point " << p << " of order " << ni << endl;
	  cout<<"generating subgroup "<<cycle<<endl;
#endif
	}
#ifdef DEBUG_TORSION
      else
	{
	  cout<<"none\n";
	}
#endif
    }

#ifdef DEBUG_TORSION
  cout<<"Number of points in cyclic part = " << ntp << endl;
  for(i=0; i<ntp; i++) cout<<points[i]<<"\t";
  cout<<endl;
#endif
  
  if(ntp==1) {return two_torsion;}                   // C2xC2
  if(nt2==4)
  // non-cyclic case, C2xC4, C2xC6 or C2xC8
  // Find a point of order 2 not in the second factor and add it in
    {
      for(i=1; i<4; i++)
	{
	  p=two_torsion[i];
	  if(points.locate(p)==-1)
	    {
#ifdef DEBUG_TORSION
	      cout<<"Using 2-torsion point "<<p<<" as coset rep\n";
#endif
	      for(j=0; j<ntp; j++)
		{
		  points[ntp+j]=points[j]+p;
		}
	      ntp*=2;
	      break;
	    }
#ifdef DEBUG_TORSION
	  cout<<"2-torsion point "<<p<<" is in subgroup\n";
#endif
	}
    }
  points.truncate(ntp);
  return points;
}

PointArray old_torsion_points(Curvedata& E)
{
  if ( E.isnull() ) return 0;
  bigint a1,a2,a3,a4,a6,sa2,sa4,sa6, d, x, y ;
  E.getai(a1,a2,a3,a4,a6);
  long i,nroots,ntp;  int scaled_flag;
  PointArray points(16);    // upper bound
  ntp = 1;
  points[0] = Point(E) ;     // zero point
  if ( (sign(a1)==0) && (sign(a3)==0) )
    {sa2=a2; sa4=a4; sa6=a6; scaled_flag=0; }
  else
    {sa2 = a1*a1 + 4*a2;
     sa4 = 8*a1*a3 + 16*a4;
     sa6 = 16*a3*a3 + 64*a6;
     scaled_flag=1; }
  d = sa2*sa2*(sa4*sa4 - 4*sa2*sa6) + 18*sa2*sa4*sa6
                    - 4*sa4*sa4*sa4 - 27*sa6*sa6;
  Point p;

// First test y=0 for points of order 2:

  bigintArray xlist = Introotscubic(sa2, sa4, sa6);
  nroots = xlist.length;
  for (i=0; i<nroots; i++)
    {
      x = xlist[i];
      if (scaled_flag) p.init(E, 2*x, - a1*x - 4*a3, 8);
      else p.init(E, x, 0);
      points[ntp++] = p;
    }

// Now test y such that y^2 divides d:

  bigintArray possible_y( sqdivs(d) );
  for( bigintvar yvar(possible_y); yvar.ok(); yvar++ )
  {
    y = yvar.value();
    xlist = Introotscubic(sa2, sa4, sa6-y*y);
    nroots = xlist.length;
    for (i=0; i<nroots; i++)
    {
      x = xlist[i];
      if (scaled_flag) p.init(E, 2*x, y - a1*x - 4*a3, 8);
      else p.init(E, x, y);
      if (order(p) > 0)
        {
          points[ntp++] = p;
          points[ntp++] = -p;  // N.B. order>2 here!
        }
    }
  }
  points.truncate(ntp);
  return points;
}

// end of file: points.cc
