//
// Points.h, adapted from Elliptic.h by OM
//

// allow for multiple includes
#ifndef _PELLIPTIC_
#define _PELLIPTIC_

#include "curve.h"

//
// class for  points on elliptic curves
//
//PointLists commented out by JC; use PointArrays instead for simplicity!
//
//class PointList; // necessary fwd declaration
class PointArray;

class Point{ 
  bigint X ; // homogeneous coordinates
  bigint Y ;
  bigint Z ;
  Curvedata *E; // pointer to the curve that the point is on
                   // warning: no check is made that client doesn't alter curve
  int ord; // 0 if not calculated yet, -1 if infinite
  bigfloat height; // -1.0 if not calculated yet, 0.0 for torsion point
  void reduce(); //divide out coordinate gcd
public:
  // constructors 
  Point(void)       // not a real point
    : E(0), X(0), Y(0), Z(0), ord(0), height(-1.0)
    { ;} 
  Point(Curvedata &EE)      // set to point at infinity
    : E(&EE), X(0), Y(1), Z(0), ord(1), height(0.0)
    { ;}      
  Point(Curvedata *EE)      // set to point at infinity
    : E(EE), X(0), Y(1), Z(0), ord(1), height(0.0)
    { ;}      
  Point(Curvedata &EE, const bigint& x, const bigint& y, const bigint& z)
    : E(&EE), X(x), Y(y), Z(z), ord(0), height(-1.0)
    { reduce(); }
  Point(Curvedata *EE, const bigint& x, const bigint& y, const bigint& z)
    : E(EE), X(x), Y(y), Z(z), ord(0), height(-1.0)
    { reduce(); }
  Point(Curvedata &EE, const bigint& x, const bigint& y)
    : E(&EE), X(x), Y(y), Z(1), ord(0), height(-1.0)
    { ;} // no need to reduce already integral
  Point(Curvedata *EE, const bigint& x, const bigint& y)
    : E(EE), X(x), Y(y), Z(1), ord(0), height(-1.0)
    { ;} // no need to reduce already integral
  Point(const Point& Q)
    :E(Q.E), X(Q.X), Y(Q.Y), Z(Q.Z), ord(Q.ord), height(Q.height)
    { ; }
  ~Point(void);
                
  // input and output
  friend inline ostream& operator<<(ostream & os, const Point& P)
    {return os << "[" << P.X << " : " << P.Y << " : " << P.Z << "]" ;}
  friend inline istream& operator>>(istream & is, Point& P)
    {char c; 
     is.flags( is.flags() | ios::dec );  //force decimal input (bug fix)
     is >> c >> P.X >> c >> P.Y >> c >> P.Z >> c;
     P.ord = 0; P.height = -1.0; if(P.isvalid()) P.reduce(); return is; }

  // test of equality of points        
  int operator==(const Point&) const;
  int operator!=(const Point&) const;
  friend int eq(const Point&P, const Point&Q);

  // assignment (p.init(.) is quicker than p=Point(.) for existing p)
  void init(Curvedata &EE,
            const bigint& x, const bigint& y, const bigint& z)
    {E=&EE; X=x; Y=y; Z=z; reduce(); ord=0; height=-1.0; }
  void init(Curvedata *EE,
            const bigint& x, const bigint& y, const bigint& z)
    {E=EE; X=x; Y=y; Z=z; reduce(); ord=0; height=-1.0; }
  void init(Curvedata &EE,
            const bigint& x, const bigint& y)
    {E=&EE; X=x; Y=y; Z = 1; ord=0; height=-1.0; }
  void init(Curvedata *EE,
            const bigint& x, const bigint& y)
    {E=EE; X=x; Y=y; Z = 1; ord=0; height=-1.0; }
  void operator=(const Point& Q) // P1 = P2
    { E=Q.E; X=Q.X ; Y=Q.Y; Z=Q.Z; ord=Q.ord; height=Q.height; }

  friend Point shift(const Point& p,  Curvedata* newc, 
		     const bigint& u=1, const bigint& r=0, 
		     const bigint& s=0, const bigint& t=0, int back=0); 

  void operator+=(const Point&) ; // P1 += P2 ; order and height unknown
  void operator-=(const Point&) ; // P1 -= P2 ; ditto

  // addition of points, etc
  Point operator+(const Point &) const ; // P1 + P2
  Point operator-(const Point &) const ; // P1 - P2
  Point operator-(void) const ; // -P
  Point twice(void) const ; // doubles P
  friend Point operator*(int, const Point&) ; // n*P
                
  // access functions
  Curve getcurve() const {return *E;}
  void getcoordinates(bigint& x, bigint& y, bigint& z) const
    {x=X; y=Y; z=Z; }
  friend inline bigint getX(const Point& p) {return p.X; }
  friend inline bigint getY(const Point& p) {return p.Y; }
  friend inline bigint getZ(const Point& p) {return p.Z; }
  friend int order(Point& p);       // calculate and set if not set
  friend int order(Point& p,  PointArray&multiples);
                        // also create and return list of multiples
  friend bigfloat height(Point& P);    //calculate and set if not set
  friend bigfloat pheight(const Point& P, const bigint& p);
  friend bigfloat realheight(const bigfloat& x, const Curvedata* E);
  friend bigfloat realheight(const Point& P);

  // useful logical tests
  int iszero() const { return Z==0; }
  int isintegral() const { return Z==1; }
  int isvalid() const ; // P on its curve ?

}; // end of point class

// the height pairing of two points
bigfloat height_pairing(Point& P, Point& Q);  

// regulator of a list of n points
bigfloat regulator(PointArray& points);  // not a const array; heights get set.

void raw_shift_point(const bigint& x, const bigint& y, const bigint& z, 
                const bigint& u, const bigint& r, 
                const bigint& s, const bigint& t, 
                bigint& newx, bigint& newy, bigint& newz, int back=0);

//
// PointArrays and Pointvars : edit of longlists, longvars
//

class PointArray {
friend class Pointvar;
  long length;        
  Point *items;  
public:
  PointArray(long n=0);
  ~PointArray();
  PointArray(const PointArray& l);
  PointArray(const PointArray& l, long n);
  PointArray& operator=(const PointArray&l);
//PointArray(PointList& ilist);
  Point& get(long i) {
    if ((i>=0)&&(i<length)) return items[i]; 
    else{cerr<<"bad index "<<i<<" in PointArray!\n"; return items[0];}}
  long getlength() const {return length;}
  long locate(const Point& num); //return index if in list, else -1
  Point& operator[](long i) {return this->get(i);}
  Point operator()(long i) const 
    {
      if ((i>=0)&&(i<length)) return items[i]; 
      else{cerr<<"bad index "<<i<<" in PointArray!\n"; return Point();}
    }
  void set(long i, const Point& val);
  void truncate(long newlength);
  void output(ostream& os) const;
  friend inline ostream& operator<<(ostream& os, const PointArray& l)
    {l.output(os);return os;}
};

class Pointvar {
public:
  long index;        /* current index */
private:
  Point* values;
  long maxindex;     /* max index */
public:
  Pointvar(const PointArray& l) 
    {maxindex=l.length-1; index=0; values=l.items;}
  void init(const PointArray& l) 
    {maxindex=l.length-1; index=0; values=l.items;}
  int next() {if ((index++)<maxindex) {values++; return 1;}
              else return 0;}
  Point operator++()
    {if ((index++)<=maxindex) {Point ans=*values;values++; return ans;}
     else return Point();}
  int ok() const {return index<=maxindex;}
  int more() const {return index<maxindex;}
  Point value() const {return *values;}
//  operator Point() const {return *values;}
};

// torsion functions (here for backwards compatibility with old versions, 
//                    redundant really)
// N.B. Don't make the params const here

PointArray torsion_points(Curvedata& E);

inline long ntorsion(Curvedata& E)
{
  return E.get_ntorsion();
}

// end of file: points.h

#endif
