// REDUCE.CC:  Reduction of quartics to minimal coeffs, same I,J
// 
#include <builtin.h>
#include "marith.h"
#include "points.h"
#include "mquartic.h"
#include "transform.h"
#include "msoluble.h"
#include "minim.h"
#include "reduce.h"

//#define REDUCE_B
//#define DEBUG_REDUCE
//#define USE_OLD_BSD  // to use bsd-reduction for type 3 (worse!)

void reduce_b(bigint& a, bigint& b, bigint& c, bigint& d, bigint& e,
	      bigint& m00, bigint& m11, bigint& m12, bigint& m21, bigint& m22)
{
  bigint a4=a<<2;
  bigint bmod4a = mod(b,a4);
  bigint alpha;
  divide_exact((bmod4a-b),a4,alpha);
  if(is_zero(alpha)) return;
  xshift(alpha,a,b,c,d,e,m00,m11,m12,m21,m22);
  return;
}

bigfloat show(bigfloat x) {cout<<x<<endl; return x;}

int one_step(const bigfloat& x0, const bigfloat& y0,
	     bigfloat& a, bigfloat& b, bigfloat& c, bigfloat& d)
// Finds a good unimodular matrix (a,b;c,d) 
// which raises z=x0+i*y0
{
  //  cout<<"In one_step with x0="<<x0<<", y0="<<y0<<endl;
  bigint k;
  bigfloat x(x0);
  bigfloat xk, x2, newc, newd, s1, s2, s;
  int i, ans=0;

  a=1; b=0; c=0; d=1;  // these are integers, but stored as bigfloats

  for(i=0; i<5; i++)
    {
      k = Iround(x); xk=I2bigfloat(k); x2=xk-x;
      //      cout<<"k = "<<k<<", xk-x = "<<x2<<endl;
      newc=xk*c-a; newd=xk*d-b;
      s1 = x0*newc+newd;
      s2 = y0*newc;
      s = s1*s1+s2*s2;
      //      cout<<"s = "<<s<<endl;
      if(s>=0.99999) return ans;
      a= c;  c= newc;  b= d;  d= newd;
      if(is_approx_zero(x2)) return ans;
      x = 1/x2;
      //      cout<<"New x = "<<x<<endl;
      ans=1;
    }
  return ans;
} 

bigfloat* types12_covar(const bigint& a, const bigint& b, const bigint& c, const bigint& d, 
			const bigfloat& xH, const bigfloat& phi)
{
  // The following covariant quadratic is the unique real quadratic factor 
  // of g6 which is positive definite
  bigfloat xa=I2bigfloat(a), xb=I2bigfloat(b), xc=I2bigfloat(c);
  bigfloat* ans = new bigfloat[3];
  ans[0] = 3*(4*xa*phi-xH);
  ans[1] = 6*(xb*phi+I2bigfloat(b*c-6*a*d));
  ans[2] = 2*phi*(xc-phi)+I2bigfloat(4*c*c-9*b*d);
#ifdef DEBUG_REDUCE
  cout<<"Before scaling, quadratic has coefficients: "
      <<ans[0]<<", "<<ans[1]<<", "<<ans[2]<<endl;
#endif
  ans[1]/=ans[0];
  ans[2]/=ans[0];
  ans[0]=1;
  return ans;
}

bigfloat* type3_covar(const bigfloat& xa, const bigfloat& xb, const bigfloat& xH, 
		      const bigfloat& rphi, const bigcomplex& cphi, int Risneg)
{
  bigfloat* ans = new bigfloat[3];
  bigfloat a4=4*xa;
  bigcomplex r1 = sqrt((a4*cphi-xH)/3);
  bigfloat r3   = sqrt((a4*rphi-xH)/3);
  if(Risneg)  r3 = -r3;
#ifdef DEBUG_REDUCE
  cout<<"r1, r3 = "<<r1<<", "<<r3<<endl;
#endif
  bigfloat rr1 = abs(real(r1));
  bigfloat ir1 = abs(imag(r1));
  bigfloat rbeta   = ( r3 - xb )        / a4;
  bigfloat ibeta   = ( 2*ir1 )          / a4;
  bigfloat alpha_1 = ( 2*rr1 - r3 - xb) / a4;
  bigfloat alpha_2 = (-2*rr1 - r3 - xb) / a4;
  // roots are alpha_1, alpha_2 (real), rbeta+/-i*ibeta (non-real)
  bigfloat p      = -2*rbeta;
  bigfloat lambda =  2*ibeta;
  bigfloat q      = (p*p+lambda*lambda)/4;
#ifdef DEBUG_REDUCE
  cout<<"phi = "<<rphi<<endl;
  cout<<"Complex phi = "<<cphi<<endl;
  cout<<"Roots are "<<alpha_1<<", "<<alpha_2<<", "<<rbeta<<" +/-i* "<<ibeta<<endl;
  cout<<"p = "<<p<<", q = "<<q<<endl;
  cout << "lambda = " << lambda << endl;
#endif
#ifdef USE_OLD_BSD
  // B&SD's covariant:
  ans[0]=1; ans[1] = p; ans[2] = q;
#else
  //
  // Julia's covariant:
  bigfloat ar = abs(r1+r3);
  bigfloat br = abs(r1-r3);
  bigfloat t1sq = ir1*ar*ar;
  bigfloat t2sq = ir1*br*br;
  bigfloat usq  = rr1*ar*br;
  
  ans[0] = t1sq + t2sq + 2*usq;
  ans[1] = -2*alpha_1*t1sq -2*alpha_2*t2sq + 2*p*usq;
  ans[2] = alpha_1*alpha_1*t1sq + alpha_2*alpha_2*t2sq + 2*q*usq;

#ifdef DEBUG_REDUCE
  cout << "quadratic has coefficients "<<ans[0]<<", "<<ans[1]<<", "<<ans[2]<<endl;
  cout << "-disc = 4*q0*q2-q1^2 = " << (4*ans[0]*ans[2]-ans[1]*ans[1]) 
       << " (should be positive!)\n";
#endif
#endif
      
  ans[1] /= ans[0];
  ans[2] /= ans[0];
  ans[0] = 1;
  
  return ans;
}

void reduce(bigint& a, bigint& b, bigint& c, bigint& d, bigint& e,
	    bigint& m00, bigint& m11, bigint& m12, bigint& m21, bigint& m22)
     // Construct a covariant quadratic, and reduce this
{
  bigint ii = II(a,b,c,d,e);
  bigint jj = JJ(a,b,c,d,e);
  bigint disc = 4*pow(ii,3)-sqr(jj);
  bigint  H = HH(a,b,c), R = RR(a,b,c,d);
  bigfloat xH = I2bigfloat(H), xR = I2bigfloat(R);
  bigfloat xii = I2bigfloat(ii), xjj=I2bigfloat(jj);
  bigcomplex c1(0), c2(-3*xii), c3(xjj);
  bigcomplex* cphi = solvecubic( c1, c2, c3);
  bigfloat * hcoeffs; // will hold coeffs of covariant quadratic
  bigfloat realphi, phi;       // will hold specific real roots
  bigfloat phi1, phi2, phi3;
  bigfloat L = (abs(cphi[0]-cphi[1]) + abs(cphi[1]-cphi[2]) + abs(cphi[2]-cphi[0]))/18;
  bigfloat a_bound;
  bigfloat xa=I2bigfloat(a), xb=I2bigfloat(b), xc=I2bigfloat(c), 
           xd=I2bigfloat(d), xe=I2bigfloat(e);
#ifdef DEBUG_REDUCE
      cout<<"Coeffs as floats are ("<<xa<<", "<<xb<<", "<<xc<<", "<<xd<<", "<<xe<<")\n";
#endif
  bigfloat a2=2*xa, a4=4*xa;;
  long type;
  bigfloat p, pdash, q, qdash, lambda, lambdadash;
  
  if(disc>0)
    {
      // all the phi are real;  order them so that a*phi[i] decreases
      bigfloat phi1 = real(cphi[0]);
      bigfloat phi2 = real(cphi[1]);
      bigfloat phi3 = real(cphi[2]);
      if(a>0)      orderreal(phi1,phi2,phi3); 
      else         orderreal(phi3,phi2,phi1); 
#ifdef DEBUG_REDUCE
      cout<<"Three real phi are "<<phi1<<", "<<phi2<<", "<<phi3<<"\n";
#endif
      // So now a*phi1>a*phi2>a*phi3
      
      if((H<0)&&(H*H>16*a*a*ii)) 
	{ 
	  type = 2; 
	  realphi = phi2; phi = phi2;
#ifdef DEBUG_REDUCE
	  cout<<"Type = 2, phi = "<<phi<<"\n";
#endif
	}
      else
	{ 
	  type = 1; 
	  realphi = phi1; phi = phi3;
#ifdef DEBUG_REDUCE
	  cout<<"Type = 1, phi = "<<phi<<"\n";
#endif
	}
      hcoeffs = types12_covar(a,b,c,d,xH,phi);
    }
  else // disc < 0
    {
      type=3;
#ifdef DEBUG_REDUCE
      cout<<"Type = 3\n";
#endif
      if (is_real(cphi[1])) 
	{
	  realphi=real(cphi[1]);
	  cphi[1]=cphi[2];
	  cphi[2]=realphi;
	}
      else 
	if (is_real(cphi[2])) 
	  {
	    realphi=real(cphi[2]);
	  }
	else 
	  {
	    realphi=real(cphi[0]);
	    cphi[0]=cphi[2];
	    cphi[2]=realphi;
	  }
      hcoeffs = type3_covar(xa,xb,xH,realphi,cphi[0],(sign(R)<0));
    }

  bigfloat xz = -hcoeffs[1]/2;
  bigfloat dz = 4*hcoeffs[2]-hcoeffs[1]*hcoeffs[1];  // should be positive!
  bigfloat az = hcoeffs[2];
  bigfloat yz = sqrt(abs(dz))/2;
  bigcomplex z(xz,yz);
#ifdef DEBUG_REDUCE
  cout << "Before reduction  "; 
  cout << "Covariant quadratic has coefficients "<<hcoeffs[0]<<", "<<hcoeffs[1]<<", "<<hcoeffs[2]<<endl;
  //  cout << "4*q2 = " << 4*hcoeffs[2] << endl;
  //  cout << "q1^2 = " << hcoeffs[1]*hcoeffs[1] << endl;
  cout << "-disc = 4*q2-q1^2 = " << dz << " (should be positive!)\n";
  cout << "Root = " << z << " with modulus^2 = " << az << endl;
#endif

  // Now do reduction.
  // nij will hold the cumulative transformation (n00=1)

  bigint n00(1), n11(1), n12(0), n21(0), n22(1); // n00 not used except for temps
  bigfloat a11, a12, a21, a22;
  bigint   na11, na12, na21, na22;
  bigint   nb11, nb12, nb21, nb22;
  int reduced=!one_step(xz, yz, a11, a12, a21, a22);;

  while(!reduced)
    {
      z = (a11*z+a12) / (a21*z+a22);
      xz=real(z); yz=imag(z);
      az  = xz*xz + yz*yz;
      
      na11=Iround(a22); na12=Iround(-a12); 
      na21=Iround(-a21); na22=Iround(a11);  // rounding the inverse matrix
#ifdef DEBUG_REDUCE
      cout << "one_step returns 0 and matrix ("<<na22<<","<<-na12<<";"<<-na21<<","<<na11<<")\n"; 
#endif

      nb11 = n11*na11 + n12*na21;
      nb12 = n11*na12 + n12*na22;
      nb21 = n21*na11 + n22*na21;
      nb22 = n21*na12 + n22*na22;
      n11 = nb11; n12 = nb12;
      n21 = nb21; n22 = nb22;

#ifdef DEBUG_REDUCE
      cout << "New root = " << z << " with modulus^2 = " << az << endl;
#endif
      reduced=!one_step(xz, yz, a11, a12, a21, a22);;
    }

  // final shift of x-coord

  bigint s = Iround(xz);
  bigfloat xs = I2bigfloat(s);
  xz -= xs;
  az  = xz*xz + yz*yz;
  n12 += s*n11;
  n22 += s*n21;

#ifdef DEBUG_REDUCE
  cout << "After reduction, root = " << z << " with modulus^2 = " << az << endl;
#endif

#ifdef DEBUG_REDUCE
  show_transform(1,n11,n12,n21,n22);
  cout<<"Now using that to further transform the quartic..."<<endl;
#endif

  apply_transform(a,b,c,d,e,1,n11,n12,n21,n22);

#ifdef DEBUG_REDUCE
  quartic newg(a,b,c,d,e);
  cout << "reduced quartic has coefficients " << newg << endl;
#endif

#ifdef REDUCE_B
  // Now reduce b so -2a < b <= 2a:
  bigint newa4=a<<2;
  bigint bmod4a = mod(b,newa4);
  bigint alpha;
  divide_exact((bmod4a-b),newa4,alpha);
  if(!is_zero(alpha))
    {
      apply_transform(a,b,c,d,e,1,1,alpha,0,1);
      n12 += alpha*n11;
      n22 += alpha*n21;
#ifdef DEBUG_REDUCE
      newg.assign(a,b,c,d,e);
      cout << "after reducing b, reduced quartic has coefficients " << newg << endl;
#endif
    }
#endif

  // Now combine the new transformation with the given one:
  n00 = m11*n11 + m12*n21;
  m12 = m11*n12 + m12*n22;
  m11 = n00;
  n00 = m21*n11 + m22*n21;
  m22 = m21*n12 + m22*n22;
  m21 = n00;

  delete [] cphi;
  return;
}

